"""Attributes of Components and properties."""

from __future__ import annotations

import itertools
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, List, Literal, Optional, Sequence, Union

from icalendar.enums import BUSYTYPE, CLASS, STATUS, TRANSP, StrEnum
from icalendar.error import IncompleteComponent, InvalidCalendar
from icalendar.parser_tools import SEQUENCE_TYPES
from icalendar.prop import (
    vCalAddress,
    vCategory,
    vDDDTypes,
    vDuration,
    vRecur,
    vText,
    vUid,
    vUri,
    vXmlReference,
)
from icalendar.prop.conference import Conference
from icalendar.prop.image import Image
from icalendar.timezone import tzp
from icalendar.tools import is_date

if TYPE_CHECKING:
    from icalendar.cal import Component

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias


def _get_rdates(
    self: Component,
) -> list[Union[tuple[date, None], tuple[datetime, None], tuple[datetime, datetime]]]:
    """The RDATE property defines the list of DATE-TIME values for recurring components.

    RDATE is defined in :rfc:`5545`.
    The return value is a list of tuples ``(start, end)``.

    ``start`` can be a :class:`datetime.date` or a :class:`datetime.datetime`,
    with and without timezone.

    ``end`` is :obj:`None` if the end is not specified and a :class:`datetime.datetime`
    if the end is specified.

    Value Type:
        The default value type for this property is DATE-TIME.
        The value type can be set to DATE or PERIOD.

    Property Parameters:
        IANA, non-standard, value data type, and time
        zone identifier property parameters can be specified on this
        property.

    Conformance:
        This property can be specified in recurring "VEVENT",
        "VTODO", and "VJOURNAL" calendar components as well as in the
        "STANDARD" and "DAYLIGHT" sub-components of the "VTIMEZONE"
        calendar component.

    Description:
        This property can appear along with the "RRULE"
        property to define an aggregate set of repeating occurrences.
        When they both appear in a recurring component, the recurrence
        instances are defined by the union of occurrences defined by both
        the "RDATE" and "RRULE".

        The recurrence dates, if specified, are used in computing the
        recurrence set.  The recurrence set is the complete set of
        recurrence instances for a calendar component.  The recurrence set
        is generated by considering the initial "DTSTART" property along
        with the "RRULE", "RDATE", and "EXDATE" properties contained
        within the recurring component.  The "DTSTART" property defines
        the first instance in the recurrence set.  The "DTSTART" property
        value SHOULD match the pattern of the recurrence rule, if
        specified.  The recurrence set generated with a "DTSTART" property
        value that doesn't match the pattern of the rule is undefined.
        The final recurrence set is generated by gathering all of the
        start DATE-TIME values generated by any of the specified "RRULE"
        and "RDATE" properties, and then excluding any start DATE-TIME
        values specified by "EXDATE" properties.  This implies that start
        DATE-TIME values specified by "EXDATE" properties take precedence
        over those specified by inclusion properties (i.e., "RDATE" and
        "RRULE").  Where duplicate instances are generated by the "RRULE"
        and "RDATE" properties, only one recurrence is considered.
        Duplicate instances are ignored.

    Example:
        Below, we set one RDATE in a list and get the resulting tuple of start and end.

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> from datetime import datetime
            >>> event = Event()

            # Add a list of recurrence dates
            >>> event.add("RDATE", [datetime(2025, 4, 28, 16, 5)])
            >>> event.rdates
            [(datetime.datetime(2025, 4, 28, 16, 5), None)]

    .. note::

        You cannot modify the RDATE value by modifying the result.
        Use :func:`icalendar.cal.Component.add` to add values.

        If you want to compute recurrences, have a look at
        `Related Projects <https://github.com/collective/icalendar/blob/main/README.rst#related-projects>`_.

    """
    result = []
    rdates = self.get("RDATE", [])
    for rdates in (rdates,) if not isinstance(rdates, list) else rdates:
        for dts in rdates.dts:
            rdate = dts.dt
            if isinstance(rdate, tuple):
                # we have a period as rdate
                if isinstance(rdate[1], timedelta):
                    result.append((rdate[0], rdate[0] + rdate[1]))
                else:
                    result.append(rdate)
            else:
                # we have a date/datetime
                result.append((rdate, None))
    return result


rdates_property = property(_get_rdates)


def _get_exdates(self: Component) -> list[date | datetime]:
    """EXDATE defines the list of DATE-TIME exceptions for recurring components.

    EXDATE is defined in :rfc:`5545`.

    Value Type:
        The default value type for this property is DATE-TIME.
        The value type can be set to DATE.

    Property Parameters:
        IANA, non-standard, value data type, and time
        zone identifier property parameters can be specified on this
        property.

    Conformance:
        This property can be specified in recurring "VEVENT",
        "VTODO", and "VJOURNAL" calendar components as well as in the
        "STANDARD" and "DAYLIGHT" sub-components of the "VTIMEZONE"
        calendar component.

    Description:
        The exception dates, if specified, are used in
        computing the recurrence set.  The recurrence set is the complete
        set of recurrence instances for a calendar component.  The
        recurrence set is generated by considering the initial "DTSTART"
        property along with the "RRULE", "RDATE", and "EXDATE" properties
        contained within the recurring component.  The "DTSTART" property
        defines the first instance in the recurrence set.  The "DTSTART"
        property value SHOULD match the pattern of the recurrence rule, if
        specified.  The recurrence set generated with a "DTSTART" property
        value that doesn't match the pattern of the rule is undefined.
        The final recurrence set is generated by gathering all of the
        start DATE-TIME values generated by any of the specified "RRULE"
        and "RDATE" properties, and then excluding any start DATE-TIME
        values specified by "EXDATE" properties.  This implies that start
        DATE-TIME values specified by "EXDATE" properties take precedence
        over those specified by inclusion properties (i.e., "RDATE" and
        "RRULE").  When duplicate instances are generated by the "RRULE"
        and "RDATE" properties, only one recurrence is considered.
        Duplicate instances are ignored.

        The "EXDATE" property can be used to exclude the value specified
        in "DTSTART".  However, in such cases, the original "DTSTART" date
        MUST still be maintained by the calendaring and scheduling system
        because the original "DTSTART" value has inherent usage
        dependencies by other properties such as the "RECURRENCE-ID".

    Example:
        Below, we add an exdate in a list and get the resulting list of exdates.

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> from datetime import datetime
            >>> event = Event()

            # Add a list of excluded dates
            >>> event.add("EXDATE", [datetime(2025, 4, 28, 16, 5)])
            >>> event.exdates
            [datetime.datetime(2025, 4, 28, 16, 5)]

    .. note::

        You cannot modify the EXDATE value by modifying the result.
        Use :func:`icalendar.cal.Component.add` to add values.

        If you want to compute recurrences, have a look at
        `Related Projects <https://github.com/collective/icalendar/blob/main/README.rst#related-projects>`_.

    """
    result = []
    exdates = self.get("EXDATE", [])
    for exdates in (exdates,) if not isinstance(exdates, list) else exdates:
        for dts in exdates.dts:
            exdate = dts.dt
            # we have a date/datetime
            result.append(exdate)
    return result


exdates_property = property(_get_exdates)


def _get_rrules(self: Component) -> list[vRecur]:
    """RRULE defines a rule or repeating pattern for recurring components.

    RRULE is defined in :rfc:`5545`.
    :rfc:`7529` adds the ``SKIP`` parameter :class:`icalendar.prop.vSkip`.

    Property Parameters:
        IANA and non-standard property parameters can
        be specified on this property.

    Conformance:
        This property can be specified in recurring "VEVENT",
        "VTODO", and "VJOURNAL" calendar components as well as in the
        "STANDARD" and "DAYLIGHT" sub-components of the "VTIMEZONE"
        calendar component, but it SHOULD NOT be specified more than once.
        The recurrence set generated with multiple "RRULE" properties is
        undefined.

    Description:
        The recurrence rule, if specified, is used in computing
        the recurrence set.  The recurrence set is the complete set of
        recurrence instances for a calendar component.  The recurrence set
        is generated by considering the initial "DTSTART" property along
        with the "RRULE", "RDATE", and "EXDATE" properties contained
        within the recurring component.  The "DTSTART" property defines
        the first instance in the recurrence set.  The "DTSTART" property
        value SHOULD be synchronized with the recurrence rule, if
        specified.  The recurrence set generated with a "DTSTART" property
        value not synchronized with the recurrence rule is undefined.  The
        final recurrence set is generated by gathering all of the start
        DATE-TIME values generated by any of the specified "RRULE" and
        "RDATE" properties, and then excluding any start DATE-TIME values
        specified by "EXDATE" properties.  This implies that start DATE-
        TIME values specified by "EXDATE" properties take precedence over
        those specified by inclusion properties (i.e., "RDATE" and
        "RRULE").  Where duplicate instances are generated by the "RRULE"
        and "RDATE" properties, only one recurrence is considered.
        Duplicate instances are ignored.

        The "DTSTART" property specified within the iCalendar object
        defines the first instance of the recurrence.  In most cases, a
        "DTSTART" property of DATE-TIME value type used with a recurrence
        rule, should be specified as a date with local time and time zone
        reference to make sure all the recurrence instances start at the
        same local time regardless of time zone changes.

        If the duration of the recurring component is specified with the
        "DTEND" or "DUE" property, then the same exact duration will apply
        to all the members of the generated recurrence set.  Else, if the
        duration of the recurring component is specified with the
        "DURATION" property, then the same nominal duration will apply to
        all the members of the generated recurrence set and the exact
        duration of each recurrence instance will depend on its specific
        start time.  For example, recurrence instances of a nominal
        duration of one day will have an exact duration of more or less
        than 24 hours on a day where a time zone shift occurs.  The
        duration of a specific recurrence may be modified in an exception
        component or simply by using an "RDATE" property of PERIOD value
        type.

    Examples:
        Daily for 10 occurrences:

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> from datetime import datetime
            >>> from zoneinfo import ZoneInfo
            >>> event = Event()
            >>> event.start = datetime(1997, 9, 2, 9, 0, tzinfo=ZoneInfo("America/New_York"))
            >>> event.add("RRULE", "FREQ=DAILY;COUNT=10")
            >>> print(event.to_ical())
            BEGIN:VEVENT
            DTSTART;TZID=America/New_York:19970902T090000
            RRULE:FREQ=DAILY;COUNT=10
            END:VEVENT
            >>> event.rrules
            [vRecur({'FREQ': ['DAILY'], 'COUNT': [10]})]

        Daily until December 24, 1997:

        .. code-block:: pycon

            >>> from icalendar import Event, vRecur
            >>> from datetime import datetime
            >>> from zoneinfo import ZoneInfo
            >>> event = Event()
            >>> event.start = datetime(1997, 9, 2, 9, 0, tzinfo=ZoneInfo("America/New_York"))
            >>> event.add("RRULE", vRecur({"FREQ": ["DAILY"]}, until=datetime(1997, 12, 24, tzinfo=ZoneInfo("UTC"))))
            >>> print(event.to_ical())
            BEGIN:VEVENT
            DTSTART;TZID=America/New_York:19970902T090000
            RRULE:FREQ=DAILY;UNTIL=19971224T000000Z
            END:VEVENT
            >>> event.rrules
            [vRecur({'FREQ': ['DAILY'], 'UNTIL': [datetime.datetime(1997, 12, 24, 0, 0, tzinfo=ZoneInfo(key='UTC'))]})]

    .. note::

        You cannot modify the RRULE value by modifying the result.
        Use :func:`icalendar.cal.Component.add` to add values.

        If you want to compute recurrences, have a look at
        `Related Projects <https://github.com/collective/icalendar/blob/main/README.rst#related-projects>`_.

    """  # noqa: E501
    rrules = self.get("RRULE", [])
    if not isinstance(rrules, list):
        return [rrules]
    return rrules


rrules_property = property(_get_rrules)


def multi_language_text_property(
    main_prop: str, compatibility_prop: Optional[str], doc: str
) -> property:
    """This creates a text property.

    This property can be defined several times with different ``LANGUAGE`` parameters.

    Args:
        main_prop (str): The property to set and get, such as ``NAME``
        compatibility_prop (str): An old property used before, such as ``X-WR-CALNAME``
        doc (str): The documentation string
    """

    def fget(self: Component) -> Optional[str]:
        """Get the property"""
        result = self.get(main_prop)
        if result is None and compatibility_prop is not None:
            result = self.get(compatibility_prop)
        if isinstance(result, list):
            for item in result:
                if "LANGUAGE" not in item.params:
                    return item
        return result

    def fset(self: Component, value: Optional[str]):
        """Set the property."""
        fdel(self)
        if value is not None:
            self.add(main_prop, value)

    def fdel(self: Component):
        """Delete the property."""
        self.pop(main_prop, None)
        if compatibility_prop is not None:
            self.pop(compatibility_prop, None)

    return property(fget, fset, fdel, doc)


def single_int_property(prop: str, default: int, doc: str) -> property:
    """Create a property for an int value that exists only once.

    Args:
        prop: The name of the property
        default: The default value
        doc: The documentation string
    """

    def fget(self: Component) -> int:
        """Get the property"""
        try:
            return int(self.get(prop, default))
        except ValueError as e:
            raise InvalidCalendar(f"{prop} must be an int") from e

    def fset(self: Component, value: Optional[int]):
        """Set the property."""
        fdel(self)
        if value is not None:
            self.add(prop, value)

    def fdel(self: Component):
        """Delete the property."""
        self.pop(prop, None)

    return property(fget, fset, fdel, doc)


def single_utc_property(name: str, docs: str) -> property:
    """Create a property to access a value of datetime in UTC timezone.

    Args:
        name: name of the property
        docs: documentation string
    """
    docs = (
        f"""The {name} property. datetime in UTC

    All values will be converted to a datetime in UTC.
    """
        + docs
    )

    def fget(self: Component) -> Optional[datetime]:
        """Get the value."""
        if name not in self:
            return None
        dt = self.get(name)
        if isinstance(dt, vText):
            # we might be in an attribute that is not typed
            value = vDDDTypes.from_ical(dt)
        else:
            value = getattr(dt, "dt", dt)
        if value is None or not isinstance(value, date):
            raise InvalidCalendar(f"{name} must be a datetime in UTC, not {value}")
        return tzp.localize_utc(value)

    def fset(self: Component, value: Optional[datetime]):
        """Set the value"""
        if value is None:
            fdel(self)
            return
        if not isinstance(value, date):
            raise TypeError(f"{name} takes a datetime in UTC, not {value}")
        fdel(self)
        self.add(name, tzp.localize_utc(value))

    def fdel(self: Component):
        """Delete the property."""
        self.pop(name, None)

    return property(fget, fset, fdel, doc=docs)


def single_string_property(
    name: str, docs: str, other_name: Optional[str] = None, default: str = ""
) -> property:
    """Create a property to access a single string value."""

    def fget(self: Component) -> str:
        """Get the value."""
        result = self.get(
            name, None if other_name is None else self.get(other_name, None)
        )
        if result is None or result == []:
            return default
        if isinstance(result, list):
            return result[0]
        return result

    def fset(self: Component, value: Optional[str]):
        """Set the value.

        Setting the value to None will delete it.
        """
        fdel(self)
        if value is not None:
            self.add(name, value)

    def fdel(self: Component):
        """Delete the property."""
        self.pop(name, None)
        if other_name is not None:
            self.pop(other_name, None)

    return property(fget, fset, fdel, doc=docs)


color_property = single_string_property(
    "COLOR",
    """This property specifies a color used for displaying the component.

    This implements :rfc:`7986` ``COLOR`` property.

    Property Parameters:
        IANA and non-standard property parameters can
        be specified on this property.

    Conformance:
        This property can be specified once in an iCalendar
        object or in ``VEVENT``, ``VTODO``, or ``VJOURNAL`` calendar components.

    Description:
        This property specifies a color that clients MAY use
        when presenting the relevant data to a user.  Typically, this
        would appear as the "background" color of events or tasks.  The
        value is a case-insensitive color name taken from the CSS3 set of
        names, defined in Section 4.3 of `W3C.REC-css3-color-20110607 <https://www.w3.org/TR/css-color-3/>`_.

    Example:
        ``"turquoise"``, ``"#ffffff"``

        .. code-block:: pycon

            >>> from icalendar import Todo
            >>> todo = Todo()
            >>> todo.color = "green"
            >>> print(todo.to_ical())
            BEGIN:VTODO
            COLOR:green
            END:VTODO
    """,
)

sequence_property = single_int_property(
    "SEQUENCE",
    0,
    """This property defines the revision sequence number of the calendar component within a sequence of revisions.

Value Type:
    INTEGER

Property Parameters:
    IANA and non-standard property parameters can be specified on this property.

Conformance:
    The property can be specified in "VEVENT", "VTODO", or
    "VJOURNAL" calendar component.

Description:
    When a calendar component is created, its sequence
    number is 0.  It is monotonically incremented by the "Organizer's"
    CUA each time the "Organizer" makes a significant revision to the
    calendar component.

    The "Organizer" includes this property in an iCalendar object that
    it sends to an "Attendee" to specify the current version of the
    calendar component.

    The "Attendee" includes this property in an iCalendar object that
    it sends to the "Organizer" to specify the version of the calendar
    component to which the "Attendee" is referring.

    A change to the sequence number is not the mechanism that an
    "Organizer" uses to request a response from the "Attendees".  The
    "RSVP" parameter on the "ATTENDEE" property is used by the
    "Organizer" to indicate that a response from the "Attendees" is
    requested.

    Recurrence instances of a recurring component MAY have different
    sequence numbers.

Examples:
    The following is an example of this property for a calendar
    component that was just created by the "Organizer":

    .. code-block:: pycon

        >>> from icalendar import Event
        >>> event = Event()
        >>> event.sequence
        0

    The following is an example of this property for a calendar
    component that has been revised 10 different times by the
    "Organizer":

    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> calendar = Calendar.example("issue_156_RDATE_with_PERIOD_TZID_khal")
        >>> event = calendar.events[0]
        >>> event.sequence
        10
    """,  # noqa: E501
)


def _get_categories(component: Component) -> list[str]:
    """Get all the categories."""
    categories: Optional[vCategory | list[vCategory]] = component.get("CATEGORIES")
    if isinstance(categories, list):
        _set_categories(
            component,
            list(itertools.chain.from_iterable(cat.cats for cat in categories)),
        )
        return _get_categories(component)
    if categories is None:
        categories = vCategory([])
        component.add("CATEGORIES", categories)
    return categories.cats


def _set_categories(component: Component, cats: Optional[Sequence[str]]) -> None:
    """Set the categories."""
    if not cats and cats != []:
        _del_categories(component)
        return
    component["CATEGORIES"] = categories = vCategory(cats)
    if isinstance(cats, list):
        cats.clear()
        cats.extend(categories.cats)
        categories.cats = cats


def _del_categories(component: Component) -> None:
    """Delete the categories."""
    component.pop("CATEGORIES", None)


categories_property = property(
    _get_categories,
    _set_categories,
    _del_categories,
    """This property defines the categories for a component.

Property Parameters:
    IANA, non-standard, and language property parameters can be specified on this
    property.

Conformance:
    The property can be specified within "VEVENT", "VTODO", or "VJOURNAL" calendar
    components.
    Since :rfc:`7986` it can also be defined on a "VCALENDAR" component.

Description:
    This property is used to specify categories or subtypes
    of the calendar component.  The categories are useful in searching
    for a calendar component of a particular type and category.
    Within the "VEVENT", "VTODO", or "VJOURNAL" calendar components,
    more than one category can be specified as a COMMA-separated list
    of categories.

Example:
    Below, we add the categories to an event:

    .. code-block:: pycon

        >>> from icalendar import Event
        >>> event = Event()
        >>> event.categories = ["Work", "Meeting"]
        >>> print(event.to_ical())
        BEGIN:VEVENT
        CATEGORIES:Work,Meeting
        END:VEVENT
        >>> event.categories.append("Lecture")
        >>> event.categories == ["Work", "Meeting", "Lecture"]
        True

.. note::

    At present, we do not take the LANGUAGE parameter into account.

.. seealso::

    :attr:`Component.concepts`
""",
)


def _get_attendees(self: Component) -> list[vCalAddress]:
    """Get attendees."""
    value = self.get("ATTENDEE")
    if value is None:
        value = []
        self["ATTENDEE"] = value
        return value
    if isinstance(value, vCalAddress):
        return [value]
    return value


def _set_attendees(self: Component, value: list[vCalAddress] | vCalAddress | None):
    """Set attendees."""
    _del_attendees(self)
    if value is None:
        return
    if not isinstance(value, list):
        value = [value]
    self["ATTENDEE"] = value


def _del_attendees(self: Component):
    """Delete all attendees."""
    self.pop("ATTENDEE", None)


attendees_property = property(
    _get_attendees,
    _set_attendees,
    _del_attendees,
    """ATTENDEE defines one or more "Attendees" within a calendar component.

Conformance:
    This property MUST be specified in an iCalendar object
    that specifies a group-scheduled calendar entity.  This property
    MUST NOT be specified in an iCalendar object when publishing the
    calendar information (e.g., NOT in an iCalendar object that
    specifies the publication of a calendar user's busy time, event,
    to-do, or journal).  This property is not specified in an
    iCalendar object that specifies only a time zone definition or
    that defines calendar components that are not group-scheduled
    components, but are components only on a single user's calendar.

Description:
    This property MUST only be specified within calendar
    components to specify participants, non-participants, and the
    chair of a group-scheduled calendar entity.  The property is
    specified within an "EMAIL" category of the "VALARM" calendar
    component to specify an email address that is to receive the email
    type of iCalendar alarm.

Examples:
    Add a new attendee to an existing event.

    .. code-block:: pycon

        >>> from icalendar import Event, vCalAddress
        >>> event = Event()
        >>> event.attendees.append(vCalAddress("mailto:me@my-domain.com"))
        >>> print(event.to_ical())
        BEGIN:VEVENT
        ATTENDEE:mailto:me@my-domain.com
        END:VEVENT

    Create an email alarm with several attendees:

        >>> from icalendar import Alarm, vCalAddress
        >>> alarm = Alarm.new(attendees = [
        ...     vCalAddress("mailto:me@my-domain.com"),
        ...     vCalAddress("mailto:you@my-domain.com"),
        ... ], summary = "Email alarm")
        >>> print(alarm.to_ical())
        BEGIN:VALARM
        ATTENDEE:mailto:me@my-domain.com
        ATTENDEE:mailto:you@my-domain.com
        SUMMARY:Email alarm
        END:VALARM
""",
)

uid_property = single_string_property(
    "UID",
    """UID specifies the persistent, globally unique identifier for a component.

We recommend using :func:`uuid.uuid4` to generate new values.

Returns:
    The value of the UID property as a string or ``""`` if no value is set.

Description:
    The "UID" itself MUST be a globally unique identifier.
    The generator of the identifier MUST guarantee that the identifier
    is unique.

    This is the method for correlating scheduling messages with the
    referenced "VEVENT", "VTODO", or "VJOURNAL" calendar component.
    The full range of calendar components specified by a recurrence
    set is referenced by referring to just the "UID" property value
    corresponding to the calendar component.  The "RECURRENCE-ID"
    property allows the reference to an individual instance within the
    recurrence set.

    This property is an important method for group-scheduling
    applications to match requests with later replies, modifications,
    or deletion requests.  Calendaring and scheduling applications
    MUST generate this property in "VEVENT", "VTODO", and "VJOURNAL"
    calendar components to assure interoperability with other group-
    scheduling applications.  This identifier is created by the
    calendar system that generates an iCalendar object.

    Implementations MUST be able to receive and persist values of at
    least 255 octets for this property, but they MUST NOT truncate
    values in the middle of a UTF-8 multi-octet sequence.

    :rfc:`7986` states that UID can be used, for
    example, to identify duplicate calendar streams that a client may
    have been given access to.  It can be used in conjunction with the
    "LAST-MODIFIED" property also specified on the "VCALENDAR" object
    to identify the most recent version of a calendar.

Conformance:
    :rfc:`5545` states that the "UID" property can be specified on "VEVENT", "VTODO",
    and "VJOURNAL" calendar components.
    :rfc:`7986` modifies the definition of the "UID" property to
    allow it to be defined in an iCalendar object.
    :rfc:`9074`  adds a "UID" property to "VALARM" components to allow a unique
    identifier to be specified. The value of this property can then be used
    to refer uniquely to the "VALARM" component.

    This property can be specified once only.

Security:
    :rfc:`7986` states that UID values MUST NOT include any data that
    might identify a user, host, domain, or any other security- or
    privacy-sensitive information.  It is RECOMMENDED that calendar user
    agents now generate "UID" values that are hex-encoded random
    Universally Unique Identifier (UUID) values as defined in
    Sections 4.4 and 4.5 of :rfc:`4122`.
    You can use the :mod:`uuid` module to generate new UUIDs.

Compatibility:
    For Alarms, ``X-ALARMUID`` is also considered.

Examples:
    The following is an example of such a property value:
    ``5FC53010-1267-4F8E-BC28-1D7AE55A7C99``.

    Set the UID of a calendar:

    .. code-block:: pycon

        >>> from icalendar import Calendar
        >>> from uuid import uuid4
        >>> calendar = Calendar()
        >>> calendar.uid = uuid4()
        >>> print(calendar.to_ical())
        BEGIN:VCALENDAR
        UID:d755cef5-2311-46ed-a0e1-6733c9e15c63
        END:VCALENDAR

""",
)

summary_property = multi_language_text_property(
    "SUMMARY",
    None,
    """SUMMARY defines a short summary or subject for the calendar component.

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

Conformance:
    The property can be specified in "VEVENT", "VTODO",
    "VJOURNAL", or "VALARM" calendar components.

Description:
    This property is used in the "VEVENT", "VTODO", and
    "VJOURNAL" calendar components to capture a short, one-line
    summary about the activity or journal entry.

    This property is used in the "VALARM" calendar component to
    capture the subject of an EMAIL category of alarm.

Examples:
    The following is an example of this property:

    .. code-block:: pycon

        SUMMARY:Department Party
""",
)

description_property = multi_language_text_property(
    "DESCRIPTION",
    None,
    """DESCRIPTION provides a more complete description of the calendar component than that provided by the "SUMMARY" property.

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

Conformance:
    The property can be specified in the "VEVENT", "VTODO",
    "VJOURNAL", or "VALARM" calendar components.  The property can be
    specified multiple times only within a "VJOURNAL" calendar
    component.

Description:
    This property is used in the "VEVENT" and "VTODO" to
    capture lengthy textual descriptions associated with the activity.

    This property is used in the "VALARM" calendar component to
    capture the display text for a DISPLAY category of alarm, and to
    capture the body text for an EMAIL category of alarm.

Examples:
    The following is an example of this property with formatted
    line breaks in the property value:

    .. code-block:: pycon

        DESCRIPTION:Meeting to provide technical review for "Phoenix"
         design.\\nHappy Face Conference Room. Phoenix design team
         MUST attend this meeting.\\nRSVP to team leader.

    """,  # noqa: E501
)


def create_single_property(
    prop: str,
    value_attr: Optional[str],
    value_type: tuple[type],
    type_def: type,
    doc: str,
    vProp: type = vDDDTypes,  # noqa: N803
):
    """Create a single property getter and setter.

    :param prop: The name of the property.
    :param value_attr: The name of the attribute to get the value from.
    :param value_type: The type of the value.
    :param type_def: The type of the property.
    :param doc: The docstring of the property.
    :param vProp: The type of the property from :mod:`icalendar.prop`.
    """

    def p_get(self: Component):
        default = object()
        result = self.get(prop, default)
        if result is default:
            return None
        if isinstance(result, list):
            raise InvalidCalendar(f"Multiple {prop} defined.")
        value = result if value_attr is None else getattr(result, value_attr, result)
        if not isinstance(value, value_type):
            raise InvalidCalendar(
                f"{prop} must be either a "
                f"{' or '.join(t.__name__ for t in value_type)},"
                f" not {value}."
            )
        return value

    def p_set(self: Component, value) -> None:
        if value is None:
            p_del(self)
            return
        if not isinstance(value, value_type):
            raise TypeError(
                f"Use {' or '.join(t.__name__ for t in value_type)}, "
                f"not {type(value).__name__}."
            )
        self[prop] = vProp(value)
        if prop in self.exclusive:
            for other_prop in self.exclusive:
                if other_prop != prop:
                    self.pop(other_prop, None)

    p_set.__annotations__["value"] = p_get.__annotations__["return"] = Optional[
        type_def
    ]

    def p_del(self: Component):
        self.pop(prop)

    p_doc = f"""The {prop} property.

    {doc}

    Accepted values: {", ".join(t.__name__ for t in value_type)}.
    If the attribute has invalid values, we raise InvalidCalendar.
    If the value is absent, we return None.
    You can also delete the value with del or by setting it to None.
    """
    return property(p_get, p_set, p_del, p_doc)


X_MOZ_SNOOZE_TIME_property = single_utc_property(
    "X-MOZ-SNOOZE-TIME", "Thunderbird: Alarms before this time are snoozed."
)
X_MOZ_LASTACK_property = single_utc_property(
    "X-MOZ-LASTACK", "Thunderbird: Alarms before this time are acknowledged."
)


def property_get_duration(self: Component) -> Optional[timedelta]:
    """Getter for property DURATION."""
    default = object()
    duration = self.get("duration", default)
    if isinstance(duration, vDDDTypes):
        return duration.dt
    if isinstance(duration, vDuration):
        return duration.td
    if duration is not default and not isinstance(duration, timedelta):
        raise InvalidCalendar(
            f"DURATION must be a timedelta, not {type(duration).__name__}."
        )
    return None


def property_set_duration(self: Component, value: Optional[timedelta]):
    """Setter for property DURATION."""
    if value is None:
        self.pop("duration", None)
        return
    if not isinstance(value, timedelta):
        raise TypeError(f"Use timedelta, not {type(value).__name__}.")
    self["duration"] = vDuration(value)
    self.pop("DTEND")
    self.pop("DUE")


def property_del_duration(self: Component):
    """Delete property DURATION."""
    self.pop("DURATION")


property_doc_duration_template = """The DURATION property.

The "DTSTART" property for a "{component}" specifies the inclusive
start of the {component}.
The "DURATION" property in conjunction with the DTSTART property
for a "{component}" calendar component specifies the non-inclusive end
of the event.

If you would like to calculate the duration of a {component}, do not use this.
Instead use the duration property (lower case).
"""


def duration_property(component: str) -> property:
    """Return the duration property."""
    return property(
        property_get_duration,
        property_set_duration,
        property_del_duration,
        property_doc_duration_template.format(component=component),
    )


def multi_text_property(name: str, docs: str) -> property:
    """Get a property that can occur several times and is text.

    Examples: Journal.descriptions, Event.comments
    """

    def fget(self: Component) -> list[str]:
        """Get the values."""
        descriptions = self.get(name)
        if descriptions is None:
            return []
        if not isinstance(descriptions, SEQUENCE_TYPES):
            return [descriptions]
        return descriptions

    def fset(self: Component, values: Optional[str | Sequence[str]]):
        """Set the values."""
        fdel(self)
        if values is None:
            return
        if isinstance(values, str):
            self.add(name, values)
        else:
            for description in values:
                self.add(name, description)

    def fdel(self: Component):
        """Delete the values."""
        self.pop(name)

    return property(fget, fset, fdel, docs)


descriptions_property = multi_text_property(
    "DESCRIPTION",
    """DESCRIPTION provides a more complete description of the calendar component than that provided by the "SUMMARY" property.

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

Conformance:
    The property can be
    specified multiple times only within a "VJOURNAL" calendar component.

Description:
    This property is used in the "VJOURNAL" calendar component to
    capture one or more textual journal entries.

Examples:
    The following is an example of this property with formatted
    line breaks in the property value:

    .. code-block:: pycon

        DESCRIPTION:Meeting to provide technical review for "Phoenix"
         design.\\nHappy Face Conference Room. Phoenix design team
         MUST attend this meeting.\\nRSVP to team leader.

""",  # noqa: E501
)

comments_property = multi_text_property(
    "COMMENT",
    """COMMENT is used to specify a comment to the calendar user.

Purpose:
    This property specifies non-processing information intended
    to provide a comment to the calendar user.

Conformance:
    In :rfc:`5545`, this property can be specified multiple times in
    "VEVENT", "VTODO", "VJOURNAL", and "VFREEBUSY" calendar components
    as well as in the "STANDARD" and "DAYLIGHT" sub-components.
    In :rfc:`7953`, this property can be specified multiple times in
    "VAVAILABILITY" and "VAVAILABLE".

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

""",
)


def _get_organizer(self: Component) -> Optional[vCalAddress]:
    """ORGANIZER defines the organizer for a calendar component.

    Property Parameters:
        IANA, non-standard, language, common name,
        directory entry reference, and sent-by property parameters can be
        specified on this property.

    Conformance:
        This property MUST be specified in an iCalendar object
        that specifies a group-scheduled calendar entity.  This property
        MUST be specified in an iCalendar object that specifies the
        publication of a calendar user's busy time.  This property MUST
        NOT be specified in an iCalendar object that specifies only a time
        zone definition or that defines calendar components that are not
        group-scheduled components, but are components only on a single
        user's calendar.

    Description:
        This property is specified within the "VEVENT",
        "VTODO", and "VJOURNAL" calendar components to specify the
        organizer of a group-scheduled calendar entity.  The property is
        specified within the "VFREEBUSY" calendar component to specify the
        calendar user requesting the free or busy time.  When publishing a
        "VFREEBUSY" calendar component, the property is used to specify
        the calendar that the published busy time came from.

        The property has the property parameters "CN", for specifying the
        common or display name associated with the "Organizer", "DIR", for
        specifying a pointer to the directory information associated with
        the "Organizer", "SENT-BY", for specifying another calendar user
        that is acting on behalf of the "Organizer".  The non-standard
        parameters may also be specified on this property.  If the
        "LANGUAGE" property parameter is specified, the identified
        language applies to the "CN" parameter value.
    """
    return self.get("ORGANIZER")


def _set_organizer(self: Component, value: Optional[vCalAddress | str]):
    """Set the value."""
    _del_organizer(self)
    if value is not None:
        self.add("ORGANIZER", value)


def _del_organizer(self: Component):
    """Delete the value."""
    self.pop("ORGANIZER")


organizer_property = property(_get_organizer, _set_organizer, _del_organizer)


def single_string_enum_property(
    name: str, enum: type[StrEnum], default: StrEnum, docs: str
) -> property:
    """Create a property to access a single string value and convert it to an enum."""
    prop = single_string_property(name, docs, default=default)

    def fget(self: Component) -> StrEnum:
        """Get the value."""
        value = prop.fget(self)
        if value == default:
            return default
        return enum(str(value))

    def fset(self: Component, value: str | StrEnum | None) -> None:
        """Set the value."""
        if value == "":
            value = None
        prop.fset(self, value)

    return property(fget, fset, prop.fdel, doc=docs)


busy_type_property = single_string_enum_property(
    "BUSYTYPE",
    BUSYTYPE,
    BUSYTYPE.BUSY_UNAVAILABLE,
    """BUSYTYPE specifies the default busy time type.

Returns:
    :class:`icalendar.enums.BUSYTYPE`

Description:
    This property is used to specify the default busy time
    type. The values correspond to those used by the FBTYPE"
    parameter used on a "FREEBUSY" property, with the exception that
    the "FREE" value is not used in this property.  If not specified
    on a component that allows this property, the default is "BUSY-
    UNAVAILABLE".
""",
)

priority_property = single_int_property(
    "PRIORITY",
    0,
    """

Conformance:
    This property can be specified in "VEVENT" and "VTODO" calendar components
    according to :rfc:`5545`.
    :rfc:`7953` adds this property to "VAVAILABILITY".

Description:
    This priority is specified as an integer in the range 0
    to 9.  A value of 0 specifies an undefined priority.  A value of 1
    is the highest priority.  A value of 2 is the second highest
    priority.  Subsequent numbers specify a decreasing ordinal
    priority.  A value of 9 is the lowest priority.

    A CUA with a three-level priority scheme of "HIGH", "MEDIUM", and
    "LOW" is mapped into this property such that a property value in
    the range of 1 to 4 specifies "HIGH" priority.  A value of 5 is
    the normal or "MEDIUM" priority.  A value in the range of 6 to 9
    is "LOW" priority.

    A CUA with a priority schema of "A1", "A2", "A3", "B1", "B2", ...,
    "C3" is mapped into this property such that a property value of 1
    specifies "A1", a property value of 2 specifies "A2", a property
    value of 3 specifies "A3", and so forth up to a property value of
    9 specifies "C3".

    Other integer values are reserved for future use.

    Within a "VEVENT" calendar component, this property specifies a
    priority for the event.  This property may be useful when more
    than one event is scheduled for a given time period.

    Within a "VTODO" calendar component, this property specified a
    priority for the to-do.  This property is useful in prioritizing
    multiple action items for a given time period.
""",
)

class_property = single_string_enum_property(
    "CLASS",
    CLASS,
    CLASS.PUBLIC,
    """CLASS specifies the class of the calendar component.

Returns:
    :class:`icalendar.enums.CLASS`

Description:
    An access classification is only one component of the
    general security system within a calendar application.  It
    provides a method of capturing the scope of the access the
    calendar owner intends for information within an individual
    calendar entry.  The access classification of an individual
    iCalendar component is useful when measured along with the other
    security components of a calendar system (e.g., calendar user
    authentication, authorization, access rights, access role, etc.).
    Hence, the semantics of the individual access classifications
    cannot be completely defined by this memo alone.  Additionally,
    due to the "blind" nature of most exchange processes using this
    memo, these access classifications cannot serve as an enforcement
    statement for a system receiving an iCalendar object.  Rather,
    they provide a method for capturing the intention of the calendar
    owner for the access to the calendar component.  If not specified
    in a component that allows this property, the default value is
    PUBLIC.  Applications MUST treat x-name and iana-token values they
    don't recognize the same way as they would the PRIVATE value.
""",
)

transparency_property = single_string_enum_property(
    "TRANSP",
    TRANSP,
    TRANSP.OPAQUE,
    """TRANSP defines whether or not an event is transparent to busy time searches.

Returns:
    :class:`icalendar.enums.TRANSP`

Description:
    Time Transparency is the characteristic of an event
    that determines whether it appears to consume time on a calendar.
    Events that consume actual time for the individual or resource
    associated with the calendar SHOULD be recorded as OPAQUE,
    allowing them to be detected by free/busy time searches.  Other
    events, which do not take up the individual's (or resource's) time
    SHOULD be recorded as TRANSPARENT, making them invisible to free/
    busy time searches.
""",
)
status_property = single_string_enum_property(
    "STATUS",
    STATUS,
    "",
    """STATUS defines the overall status or confirmation for the calendar component.

Returns:
    :class:`icalendar.enums.STATUS`

The default value is ``""``.

Description:
    In a group-scheduled calendar component, the property
    is used by the "Organizer" to provide a confirmation of the event
    to the "Attendees".  For example in a "VEVENT" calendar component,
    the "Organizer" can indicate that a meeting is tentative,
    confirmed, or cancelled.  In a "VTODO" calendar component, the
    "Organizer" can indicate that an action item needs action, is
    completed, is in process or being worked on, or has been
    cancelled.  In a "VJOURNAL" calendar component, the "Organizer"
    can indicate that a journal entry is draft, final, or has been
    cancelled or removed.
""",
)

url_property = single_string_property(
    "URL",
    """A Uniform Resource Locator (URL) associated with the iCalendar object.

Description:
    This property may be used in a calendar component to
    convey a location where a more dynamic rendition of the calendar
    information associated with the calendar component can be found.
    This memo does not attempt to standardize the form of the URI, nor
    the format of the resource pointed to by the property value.  If
    the URL property and Content-Location MIME header are both
    specified, they MUST point to the same resource.

Conformance:
    This property can be specified once in the "VEVENT",
    "VTODO", "VJOURNAL", or "VFREEBUSY" calendar components.
    Since :rfc:`7986`, this property can also be defined on a "VCALENDAR".

Example:
    The following is an example of this property:

    .. code-block:: text

        URL:http://example.com/pub/calendars/jsmith/mytime.ics

""",
)

source_property = single_string_property(
    "SOURCE",
    """A URI from where calendar data can be refreshed.

Description:
    This property identifies a location where a client can
    retrieve updated data for the calendar.  Clients SHOULD honor any
    specified "REFRESH-INTERVAL" value when periodically retrieving
    data.  Note that this property differs from the "URL" property in
    that "URL" is meant to provide an alternative representation of
    the calendar data rather than the original location of the data.

Conformance:
    This property can be specified once in an iCalendar object.

Example:
    The following is an example of this property:

    .. code-block:: text

        SOURCE;VALUE=URI:https://example.com/holidays.ics

""",
)

location_property = multi_language_text_property(
    "LOCATION",
    None,
    """The intended venue for the activity defined by a calendar component.

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

Conformance:
    Since :rfc:`5545`, this property can be specified in "VEVENT" or "VTODO"
    calendar component.
    :rfc:`7953` adds this property to "VAVAILABILITY" and "VAVAILABLE".

Description:
    Specific venues such as conference or meeting rooms may
    be explicitly specified using this property.  An alternate
    representation may be specified that is a URI that points to
    directory information with more structured specification of the
    location.  For example, the alternate representation may specify
    either an LDAP URL :rfc:`4516` pointing to an LDAP server entry or a
    CID URL :rfc:`2392` pointing to a MIME body part containing a
    Virtual-Information Card (vCard) :rfc:`2426` for the location.

""",
)

contacts_property = multi_text_property(
    "CONTACT",
    """Contact information associated with the calendar component.

Purpose:
    This property is used to represent contact information or
    alternately a reference to contact information associated with the
    calendar component.

Property Parameters:
    IANA, non-standard, alternate text
    representation, and language property parameters can be specified
    on this property.

Conformance:
    In :rfc:`5545`, this property can be specified in a "VEVENT", "VTODO",
    "VJOURNAL", or "VFREEBUSY" calendar component.
    In :rfc:`7953`, this property can be specified in a "VAVAILABILITY"
    amd "VAVAILABLE" calendar component.

Description:
    The property value consists of textual contact
    information.  An alternative representation for the property value
    can also be specified that refers to a URI pointing to an
    alternate form, such as a vCard :rfc:`2426`, for the contact
    information.

Example:
    The following is an example of this property referencing
    textual contact information:

    .. code-block:: text

        CONTACT:Jim Dolittle\\, ABC Industries\\, +1-919-555-1234

    The following is an example of this property with an alternate
    representation of an LDAP URI to a directory entry containing the
    contact information:

    .. code-block:: text

        CONTACT;ALTREP="ldap://example.com:6666/o=ABC%20Industries\\,
        c=US???(cn=Jim%20Dolittle)":Jim Dolittle\\, ABC Industries\\,
        +1-919-555-1234

    The following is an example of this property with an alternate
    representation of a MIME body part containing the contact
    information, such as a vCard :rfc:`2426` embedded in a text/
    directory media type :rfc:`2425`:

    .. code-block:: text

        CONTACT;ALTREP="CID:part3.msg970930T083000SILVER@example.com":
         Jim Dolittle\\, ABC Industries\\, +1-919-555-1234

    The following is an example of this property referencing a network
    resource, such as a vCard :rfc:`2426` object containing the contact
    information:

    .. code-block:: text

        CONTACT;ALTREP="http://example.com/pdi/jdoe.vcf":Jim
         Dolittle\\, ABC Industries\\, +1-919-555-1234
""",
)


def timezone_datetime_property(name: str, docs: str):
    """Create a property to access the values with a proper timezone."""

    return single_utc_property(name, docs)


rfc_7953_dtstart_property = timezone_datetime_property(
    "DTSTART",
    """Start of the component.

    This is almost the same as :attr:`Event.DTSTART` with one exception:
    The values MUST have a timezone and DATE is not allowed.

    Description:
        :rfc:`7953`: If specified, the "DTSTART" and "DTEND" properties in
        "VAVAILABILITY" components and "AVAILABLE" subcomponents MUST be
        "DATE-TIME" values specified as either the date with UTC time or
        the date with local time and a time zone reference.

    """,
)

rfc_7953_dtend_property = timezone_datetime_property(
    "DTEND",
    """Start of the component.

    This is almost the same as :attr:`Event.DTEND` with one exception:
    The values MUST have a timezone and DATE is not allowed.

    Description:
        :rfc:`7953`: If specified, the "DTSTART" and "DTEND" properties in
        "VAVAILABILITY" components and "AVAILABLE" subcomponents MUST be
        "DATE-TIME" values specified as either the date with UTC time or
        the date with local time and a time zone reference.
    """,
)


@property
def rfc_7953_duration_property(self) -> Optional[timedelta]:
    """Compute the duration of this component.

    If there is no :attr:`DTEND` or :attr:`DURATION` set, this is None.
    Otherwise, the duration is calculated from :attr:`DTSTART` and
    :attr:`DTEND`/:attr:`DURATION`.

    This is in accordance with :rfc:`7953`:
    If "DTEND" or "DURATION" are not present, then the end time is unbounded.
    """
    duration = self.DURATION
    if duration:
        return duration
    end = self.DTEND
    if end is None:
        return None
    start = self.DTSTART
    if start is None:
        raise IncompleteComponent("Cannot compute duration without start.")
    return end - start


@property
def rfc_7953_end_property(self) -> Optional[timedelta]:
    """Compute the duration of this component.

    If there is no :attr:`DTEND` or :attr:`DURATION` set, this is None.
    Otherwise, the duration is calculated from :attr:`DTSTART` and
    :attr:`DTEND`/:attr:`DURATION`.

    This is in accordance with :rfc:`7953`:
    If "DTEND" or "DURATION" are not present, then the end time is unbounded.
    """
    duration = self.DURATION
    if duration:
        start = self.DTSTART
        if start is None:
            raise IncompleteComponent("Cannot compute end without start.")
        return start + duration
    end = self.DTEND
    if end is None:
        return None
    return end


@rfc_7953_end_property.setter
def rfc_7953_end_property(self, value: datetime):
    self.DTEND = value


@rfc_7953_end_property.deleter
def rfc_7953_end_property(self):
    del self.DTEND


def get_start_end_duration_with_validation(
    component: Component,
    start_property: str,
    end_property: str,
    component_name: str,
) -> tuple[date | datetime | None, date | datetime | None, timedelta | None]:
    """
    Validate the component and return start, end, and duration.

    This tests validity according to :rfc:`5545` rules
    for ``Event`` and ``Todo`` components.

    Args:
        component: The component to validate, either ``Event`` or ``Todo``.
        start_property: The start property name, ``DTSTART``.
        end_property: The end property name, either ``DTEND`` for ``Event`` or
            ``DUE`` for ``Todo``.
        component_name: The component name for error messages,
            either ``VEVENT`` or ``VTODO``.

    Returns:
        tuple: (start, end, duration) values from the component.

    Raises:
        InvalidCalendar: If the component violates RFC 5545 constraints.

    """
    start = getattr(component, start_property, None)
    end = getattr(component, end_property, None)
    duration = component.DURATION

    # RFC 5545: Only one of end property and DURATION may be present
    if duration is not None and end is not None:
        end_name = "DTEND" if end_property == "DTEND" else "DUE"
        msg = (
            f"Only one of {end_name} and DURATION "
            f"may be in a {component_name}, not both."
        )
        raise InvalidCalendar(msg)

    # RFC 5545: When DTSTART is a date, DURATION must be of days or weeks
    if (
        start is not None
        and is_date(start)
        and duration is not None
        and duration.seconds != 0
    ):
        msg = "When DTSTART is a date, DURATION must be of days or weeks."
        raise InvalidCalendar(msg)

    # RFC 5545: DTSTART and end property must be of the same type
    if start is not None and end is not None and is_date(start) != is_date(end):
        end_name = "DTEND" if end_property == "DTEND" else "DUE"
        msg = (
            f"DTSTART and {end_name} must be of the same type, either date or datetime."
        )
        raise InvalidCalendar(msg)

    return start, end, duration


def get_start_property(component: Component) -> date | datetime:
    """
    Get the start property with validation.

    Args:
        component: The component from which to get its start property.

    Returns:
        The ``DTSTART`` value.

    Raises:
        IncompleteComponent: If no ``DTSTART`` is present.

    """
    # Trigger validation by calling _get_start_end_duration
    start, end, duration = component._get_start_end_duration()  # noqa: SLF001
    if start is None:
        msg = "No DTSTART given."
        raise IncompleteComponent(msg)
    return start


def get_end_property(component: Component, end_property: str) -> date | datetime:
    """
    Get the end property with fallback logic for ``Event`` and ``Todo`` components.

    Args:
        component: The component to get end from
        end_property: The end property name, either ``DTEND`` for ``Event`` or
            ``DUE`` for ``Todo``.

    Returns:
        The computed end value.

    Raises:
        IncompleteComponent: If the provided information is incomplete
            to compute the end property.

    """
    # Trigger validation by calling _get_start_end_duration
    start, end, duration = component._get_start_end_duration()  # noqa: SLF001

    if end is None and duration is None:
        if start is None:
            end_name = "DTEND" if end_property == "DTEND" else "DUE"
            msg = f"No {end_name} or DURATION+DTSTART given."
            raise IncompleteComponent(msg)

        # Default behavior differs for Event vs Todo:
        # Event: date gets +1 day, datetime gets same time
        # Todo: both date and datetime get same time (issue #898)
        if end_property == "DTEND" and is_date(start):
            return start + timedelta(days=1)
        return start

    if duration is not None:
        if start is not None:
            return start + duration
        end_name = "DTEND" if end_property == "DTEND" else "DUE"
        msg = f"No {end_name} or DURATION+DTSTART given."
        raise IncompleteComponent(msg)

    return end


def get_duration_property(component: Component) -> timedelta:
    """
    Get the duration property with fallback calculation from start and end.

    Args:
        component: The component from which to get its duration property.

    Returns:
        The duration as a timedelta.

    """
    # First check if DURATION property is explicitly set
    if "DURATION" in component:
        return component["DURATION"].dt

    # Fall back to calculated duration from start and end
    return component.end - component.start


def set_duration_with_locking(
    component: Component,
    duration: timedelta | None,
    locked: Literal["start", "end"],
    end_property: str,
) -> None:
    """
    Set the duration with explicit locking behavior for ``Event`` and ``Todo``.

    Args:
        component: The component to modify, either ``Event`` or ``Todo``.
        duration: The duration to set, or ``None`` to convert to ``DURATION`` property.
        locked: Which property to keep unchanged, either ``start`` or ``end``.
        end_property: The end property name, either ``DTEND`` for ``Event`` or
            ``DUE`` for ``Todo``.

    """
    # Convert to DURATION property if duration is None
    if duration is None:
        if "DURATION" in component:
            return  # Already has DURATION property
        current_duration = component.duration
        component.DURATION = current_duration
        return

    if not isinstance(duration, timedelta):
        msg = f"Use timedelta, not {type(duration).__name__}."
        raise TypeError(msg)

    # Validate date/duration compatibility
    start = component.DTSTART
    if start is not None and is_date(start) and duration.seconds != 0:
        msg = "When DTSTART is a date, DURATION must be of days or weeks."
        raise InvalidCalendar(msg)

    if locked == "start":
        # Keep start locked, adjust end
        if start is None:
            msg = "Cannot set duration without DTSTART. Set start time first."
            raise IncompleteComponent(msg)
        component.pop(end_property, None)  # Remove end property
        component.DURATION = duration
    elif locked == "end":
        # Keep end locked, adjust start
        current_end = component.end
        component.DTSTART = current_end - duration
        component.pop(end_property, None)  # Remove end property
        component.DURATION = duration
    else:
        msg = f"locked must be 'start' or 'end', not {locked!r}"
        raise ValueError(msg)


def set_start_with_locking(
    component: Component,
    start: date | datetime,
    locked: Literal["duration", "end"] | None,
    end_property: str,
) -> None:
    """
    Set the start with explicit locking behavior for ``Event`` and ``Todo`` components.

    Args:
        component: The component to modify, either ``Event`` or ``Todo``.
        start: The start time to set.
        locked: Which property to keep unchanged, either ``duration``, ``end``,
            or ``None`` for auto-detect.
        end_property: The end property name, either ``DTEND`` for ``Event`` or
            ``DUE`` for ``Todo``.

    """
    if locked is None:
        # Auto-detect based on existing properties
        if "DURATION" in component:
            locked = "duration"
        elif end_property in component:
            locked = "end"
        else:
            # Default to duration if no existing properties
            locked = "duration"

    if locked == "duration":
        # Keep duration locked, adjust end
        current_duration = (
            component.duration
            if "DURATION" in component or end_property in component
            else None
        )
        component.DTSTART = start
        if current_duration is not None:
            component.pop(end_property, None)  # Remove end property
            component.DURATION = current_duration
    elif locked == "end":
        # Keep end locked, adjust duration
        current_end = component.end
        component.DTSTART = start
        component.pop("DURATION", None)  # Remove duration property
        setattr(component, end_property, current_end)
    else:
        msg = f"locked must be 'duration', 'end', or None, not {locked!r}"
        raise ValueError(msg)


def set_end_with_locking(
    component: Component,
    end: date | datetime,
    locked: Literal["start", "duration"],
    end_property: str,
) -> None:
    """
    Set the end with explicit locking behavior for Event and Todo components.

    Args:
        component: The component to modify, either ``Event`` or ``Todo``.
        end: The end time to set.
        locked: Which property to keep unchanged, either ``start`` or ``duration``.
        end_property: The end property name, either ``DTEND`` for ``Event`` or ``DUE``
            for ``Todo``.

    """
    if locked == "start":
        # Keep start locked, adjust duration
        component.pop("DURATION", None)  # Remove duration property
        setattr(component, end_property, end)
    elif locked == "duration":
        # Keep duration locked, adjust start
        current_duration = component.duration
        component.DTSTART = end - current_duration
        component.pop(end_property, None)  # Remove end property
        component.DURATION = current_duration
    else:
        msg = f"locked must be 'start' or 'duration', not {locked!r}"
        raise ValueError(msg)


def _get_images(self: Component) -> list[Image]:
    """IMAGE specifies an image associated with the calendar or a calendar component.

    Description:
        This property specifies an image for an iCalendar
        object or a calendar component via a URI or directly with inline
        data that can be used by calendar user agents when presenting the
        calendar data to a user.  Multiple properties MAY be used to
        specify alternative sets of images with, for example, varying
        media subtypes, resolutions, or sizes.  When multiple properties
        are present, calendar user agents SHOULD display only one of them,
        picking one that provides the most appropriate image quality, or
        display none.  The "DISPLAY" parameter is used to indicate the
        intended display mode for the image.  The "ALTREP" parameter,
        defined in :rfc:`5545`, can be used to provide a "clickable" image
        where the URI in the parameter value can be "launched" by a click
        on the image in the calendar user agent.

    Conformance:
        This property can be specified multiple times in an
        iCalendar object or in "VEVENT", "VTODO", or "VJOURNAL" calendar
        components.

    .. note::

        At the present moment, this property is read-only. If you require a setter,
        please open an issue or a pull request.
    """
    images = self.get("IMAGE", [])
    if not isinstance(images, SEQUENCE_TYPES):
        images = [images]
    return [Image.from_property_value(img) for img in images]


images_property = property(_get_images)


def _get_conferences(self: Component) -> list[Conference]:
    """Return the CONFERENCE properties as a list.

    Purpose:
        This property specifies information for accessing a conferencing system.

    Conformance:
        This property can be specified multiple times in a
        "VEVENT" or "VTODO" calendar component.

    Description:
        This property specifies information for accessing a
        conferencing system for attendees of a meeting or task.  This
        might be for a telephone-based conference number dial-in with
        access codes included (such as a tel: URI :rfc:`3966` or a sip: or
        sips: URI :rfc:`3261`), for a web-based video chat (such as an http:
        or https: URI :rfc:`7230`), or for an instant messaging group chat
        room (such as an xmpp: URI :rfc:`5122`).  If a specific URI for a
        conferencing system is not available, a data: URI :rfc:`2397`
        containing a text description can be used.

        A conference system can be a bidirectional communication channel
        or a uni-directional "broadcast feed".

        The "FEATURE" property parameter is used to describe the key
        capabilities of the conference system to allow a client to choose
        the ones that give the required level of interaction from a set of
        multiple properties.

        The "LABEL" property parameter is used to convey additional
        details on the use of the URI.  For example, the URIs or access
        codes for the moderator and attendee of a teleconference system
        could be different, and the "LABEL" property parameter could be
        used to "tag" each "CONFERENCE" property to indicate which is
        which.

        The "LANGUAGE" property parameter can be used to specify the
        language used for text values used with this property (as per
        Section 3.2.10 of :rfc:`5545`).

    Example:
        The following are examples of this property:

        .. code-block:: text

            CONFERENCE;VALUE=URI;FEATURE=PHONE,MODERATOR;
             LABEL=Moderator dial-in:tel:+1-412-555-0123,,,654321
            CONFERENCE;VALUE=URI;FEATURE=PHONE;
             LABEL=Attendee dial-in:tel:+1-412-555-0123,,,555123
            CONFERENCE;VALUE=URI;FEATURE=PHONE;
             LABEL=Attendee dial-in:tel:+1-888-555-0456,,,555123
            CONFERENCE;VALUE=URI;FEATURE=CHAT;
             LABEL=Chat room:xmpp:chat-123@conference.example.com
            CONFERENCE;VALUE=URI;FEATURE=AUDIO,VIDEO;
             LABEL=Attendee dial-in:https://chat.example.com/audio?id=123456

        Get all conferences:

        .. code-block:: pycon

            >>> from icalendar import Event
            >>> event = Event()
            >>> event.conferences
            []

        Set a conference:

        .. code-block:: pycon

            >>> from icalendar import Event, Conference
            >>> event = Event()
            >>> event.conferences = [
            ...     Conference(
            ...         "tel:+1-412-555-0123,,,654321",
            ...         feature="PHONE,MODERATOR",
            ...         label="Moderator dial-in",
            ...         language="EN",
            ...     )
            ... ]
            >>> print(event.to_ical())
            BEGIN:VEVENT
            CONFERENCE;FEATURE="PHONE,MODERATOR";LABEL=Moderator dial-in;LANGUAGE=EN;V
             ALUE=URI:tel:+1-412-555-0123,,,654321
            END:VEVENT

    """
    conferences = self.get("CONFERENCE", [])
    if not isinstance(conferences, SEQUENCE_TYPES):
        conferences = [conferences]
    return [Conference.from_uri(conference) for conference in conferences]


def _set_conferences(self: Component, conferences: list[Conference] | None):
    """Set the conferences."""
    _del_conferences(self)
    for conference in conferences or []:
        self.add("CONFERENCE", conference.to_uri())


def _del_conferences(self: Component):
    """Delete all conferences."""
    self.pop("CONFERENCE")


conferences_property = property(_get_conferences, _set_conferences, _del_conferences)


def _get_links(self: Component) -> list[vUri | vUid | vXmlReference]:
    """LINK properties as a list.

    Purpose:
        LINK provides a reference to external information related to a component.

    Property Parameters:
        The VALUE parameter is required.
        Non-standard, link relation type, format type, label, and language parameters
        can also be specified on this property.
        The LABEL parameter is defined in :rfc:`7986`.

    Conformance:
        This property can be specified zero or more times in any iCalendar component.
        LINK is specified in :rfc:`9253`.
        The LINKREL parameter is required.

    Description:
        When used in a component, the value of this property points to
        additional information related to the component.
        For example, it may reference the originating web server.

        This property is a serialization of the model in :rfc:`8288`,
        where the link target is carried in the property value,
        the link context is the containing calendar entity,
        and the link relation type and any target attributes
        are carried in iCalendar property parameters.

        The LINK property parameters map to :rfc:`8288` attributes as follows:

        LABEL
            This parameter maps to the "title"
            attribute defined in Section 3.4.1 of :rfc:`8288`.
            LABEL is used to label the destination
            of a link such that it can be used as a human-readable identifier
            (e.g., a menu entry) in the language indicated by the LANGUAGE
            (if present).
        LANGUAGE
            This parameter maps to the "hreflang" attribute defined in Section 3.4.1
            of :rfc:`8288`. See :rfc:`5646`. Example: ``en``, ``de-ch``.
        LINKREL
            This parameter maps to the link relation type defined in Section 2.1 of
            :rfc:`8288`. See `Registered Link Relation Types
            <https://www.iana.org/assignments/link-relations/link-relations.xhtml>`_.
        FMTTYPE
            This parameter maps to the "type" attribute defined in Section 3.4.1 of
            :rfc:`8288`.

        There is no mapping for "title*", "anchor", "rev", or "media" :rfc:`8288`.

    Examples:
        The following is an example of this property,
        which provides a reference to the source for the calendar object.

        .. code-block:: text

            LINK;LINKREL=SOURCE;LABEL=Venue;VALUE=URI:
             https://example.com/events

        The following is an example of this property,
        which provides a reference to an entity from which this one was derived.
        The link relation is a vendor-defined value.

        .. code-block:: text

            LINK;LINKREL="https://example.com/linkrel/derivedFrom";
             VALUE=URI:
             https://example.com/tasks/01234567-abcd1234.ics

        The following is an example of this property,
        which provides a reference to a fragment of an XML document.
        The link relation is a vendor-defined value.

        .. code-block:: text

            LINK;LINKREL="https://example.com/linkrel/costStructure";
             VALUE=XML-REFERENCE:
             https://example.com/xmlDocs/bidFramework.xml
             #xpointer(descendant::CostStruc/range-to(
             following::CostStrucEND[1]))

        Set a link :class:`icalendar.vUri` to the event page:

        .. code-block:: pycon

            >>> from icalendar import Event, vUri
            >>> from datetime import datetime
            >>> link = vUri(
            ...     "http://example.com/event-page",
            ...     params={"LINKREL":"SOURCE"}
            ... )
            >>> event = Event.new(
            ...     start=datetime(2025, 9, 17, 12, 0),
            ...     summary="An Example Event with a page"
            ... )
            >>> event.links = [link]
            >>> print(event.to_ical())
            BEGIN:VEVENT
            SUMMARY:An Example Event with a page
            DTSTART:20250917T120000
            DTSTAMP:20250517T080612Z
            UID:d755cef5-2311-46ed-a0e1-6733c9e15c63
            LINK;LINKREL="SOURCE":http://example.com/event-page
            END:VEVENT

    """
    links = self.get("LINK", [])
    if not isinstance(links, list):
        links = [links]
    return links


LINKS_TYPE_SETTER: TypeAlias = Union[
    str, vUri, vUid, vXmlReference, None, List[Union[str, vUri, vUid, vXmlReference]]
]


def _set_links(self: Component, links: LINKS_TYPE_SETTER) -> None:
    """Set the LINKs."""
    _del_links(self)
    if links is None:
        return
    if isinstance(links, (str, vUri, vUid, vXmlReference)):
        links = [links]
    for link in links:
        if type(link) is str:
            link = vUri(link, params={"VALUE": "URI"})  # noqa: PLW2901
        self.add("LINK", link)


def _del_links(self: Component) -> None:
    """Delete all links."""
    self.pop("LINK")


links_property = property(_get_links, _set_links, _del_links)

RELATED_TO_TYPE_SETTER: TypeAlias = Union[
    None, str, vText, vUri, vUid, List[Union[str, vText, vUri, vUid]]
]


def _get_related_to(self: Component) -> list[Union[vText, vUri, vUid]]:
    """RELATED-TO properties as a list.

    Purpose:
        This property is used to represent a relationship or reference
        between one calendar component and another.
        :rfc:`9523` allows URI or UID values and a GAP parameter.

    Value Type:
        :rfc:`5545`: TEXT
        :rfc:`9253`: URI, UID

    Conformance:
        Since :rfc:`5545`. this property can be specified in the "VEVENT",
        "VTODO", and "VJOURNAL" calendar components.
        Since :rfc:`9523`, this property MAY be specified in any
        iCalendar component.

    Description (:rfc:`5545`):
        The property value consists of the persistent, globally
        unique identifier of another calendar component.  This value would
        be represented in a calendar component by the "UID" property.

        By default, the property value points to another calendar
        component that has a PARENT relationship to the referencing
        object.  The "RELTYPE" property parameter is used to either
        explicitly state the default PARENT relationship type to the
        referenced calendar component or to override the default PARENT
        relationship type and specify either a CHILD or SIBLING
        relationship.  The PARENT relationship indicates that the calendar
        component is a subordinate of the referenced calendar component.
        The CHILD relationship indicates that the calendar component is a
        superior of the referenced calendar component.  The SIBLING
        relationship indicates that the calendar component is a peer of
        the referenced calendar component.

        Changes to a calendar component referenced by this property can
        have an implicit impact on the related calendar component.  For
        example, if a group event changes its start or end date or time,
        then the related, dependent events will need to have their start
        and end dates changed in a corresponding way.  Similarly, if a
        PARENT calendar component is cancelled or deleted, then there is
        an implied impact to the related CHILD calendar components.  This
        property is intended only to provide information on the
        relationship of calendar components.  It is up to the target
        calendar system to maintain any property implications of this
        relationship.

    Description (:rfc:`9253`):
        By default or when VALUE=UID is specified, the property value
        consists of the persistent, globally unique identifier of another
        calendar component. This value would be represented in a calendar
        component by the UID property.

        By default, the property value
        points to another calendar component that has a PARENT relationship
        to the referencing object. The RELTYPE property parameter is used
        to either explicitly state the default PARENT relationship type to
        the referenced calendar component or to override the default
        PARENT relationship type and specify either a CHILD or SIBLING
        relationship or a temporal relationship.

        The PARENT relationship
        indicates that the calendar component is a subordinate of the
        referenced calendar component. The CHILD relationship indicates
        that the calendar component is a superior of the referenced calendar
        component. The SIBLING relationship indicates that the calendar
        component is a peer of the referenced calendar component.

        To preserve backwards compatibility, the value type MUST
        be UID when the PARENT, SIBLING, or CHILD relationships
        are specified.

        The FINISHTOSTART, FINISHTOFINISH, STARTTOFINISH,
        or STARTTOSTART relationships define temporal relationships, as
        specified in the RELTYPE parameter definition.

        The FIRST and NEXT
        define ordering relationships between calendar components.

        The DEPENDS-ON relationship indicates that the current calendar
        component depends on the referenced calendar component in some manner.
        For example, a task may be blocked waiting on the other,
        referenced, task.

        The REFID and CONCEPT relationships establish
        a reference from the current component to the referenced component.
        Changes to a calendar component referenced by this property
        can have an implicit impact on the related calendar component.
        For example, if a group event changes its start or end date or
        time, then the related, dependent events will need to have their
        start and end dates and times changed in a corresponding way.
        Similarly, if a PARENT calendar component is canceled or deleted,
        then there is an implied impact to the related CHILD calendar
        components. This property is intended only to provide information
        on the relationship of calendar components.

        Deletion of the target component, for example, the target of a
        FIRST, NEXT, or temporal relationship, can result in broken links.

        It is up to the target calendar system to maintain any property
        implications of these relationships.

    Examples:
        :rfc:`5545` examples of this property:

        .. code-block:: text

            RELATED-TO:jsmith.part7.19960817T083000.xyzMail@example.com

        .. code-block:: text

            RELATED-TO:19960401-080045-4000F192713-0052@example.com

        :rfc:`9253` examples of this property:

        .. code-block:: text

            RELATED-TO;VALUE=URI;RELTYPE=STARTTOFINISH:
             https://example.com/caldav/user/jb/cal/
             19960401-080045-4000F192713.ics

    See also :class:`icalendar.enum.RELTYPE`.

    """
    result = self.get("RELATED-TO", [])
    if not isinstance(result, list):
        return [result]
    return result


def _set_related_to(self: Component, values: RELATED_TO_TYPE_SETTER) -> None:
    """Set the RELATED-TO properties."""
    _del_related_to(self)
    if values is None:
        return
    if not isinstance(values, list):
        values = [values]
    for value in values:
        self.add("RELATED-TO", value)


def _del_related_to(self: Component):
    """Delete the RELATED-TO properties."""
    self.pop("RELATED-TO", None)


related_to_property = property(_get_related_to, _set_related_to, _del_related_to)


def _get_concepts(self: Component) -> list[vUri]:
    """CONCEPT

    Purpose:
        CONCEPT defines the formal categories for a calendar component.

    Conformance:
        Since :rfc:`9253`,
        this property can be specified zero or more times in any iCalendar component.

    Description:
        This property is used to specify formal categories or classifications of
        the calendar component. The values are useful in searching for a calendar
        component of a particular type and category.

        This categorization is distinct from the more informal "tagging" of components
        provided by the existing CATEGORIES property. It is expected that the value of
        the CONCEPT property will reference an external resource that provides
        information about the categorization.

        In addition, a structured URI value allows for hierarchical categorization of
        events.

        Possible category resources are the various proprietary systems, for example,
        the Library of Congress, or an open source of categorization data.

    Examples:
        The following is an example of this property.
        It points to a server acting as the source for the calendar object.

        .. code-block:: text

            CONCEPT:https://example.com/event-types/arts/music

    .. seealso::

        :attr:`Component.categories`
    """
    concepts = self.get("CONCEPT", [])
    if not isinstance(concepts, list):
        concepts = [concepts]
    return concepts


CONCEPTS_TYPE_SETTER: TypeAlias = Union[List[Union[vUri, str]], str, vUri, None]


def _set_concepts(self: Component, concepts: CONCEPTS_TYPE_SETTER):
    """Set the concepts."""
    _del_concepts(self)
    if concepts is None:
        return
    if not isinstance(concepts, list):
        concepts = [concepts]
    for value in concepts:
        self.add("CONCEPT", value)


def _del_concepts(self: Component):
    """Delete the concepts."""
    self.pop("CONCEPT", None)


concepts_property = property(_get_concepts, _set_concepts, _del_concepts)


def multi_string_property(name: str, doc: str):
    """A property for an iCalendar Property that can occur multiple times."""

    def fget(self: Component) -> list[str]:
        """Get the values of a multi-string property."""
        value = self.get(name, [])
        if not isinstance(value, list):
            value = [value]
        return value

    def fset(self: Component, value: list[str] | str | None) -> None:
        """Set the values of a multi-string property."""
        fdel(self)
        if value is None:
            return
        if not isinstance(value, list):
            value = [value]
        for value in value:
            self.add(name, value)

    def fdel(self: Component):
        """Delete the values of a multi-string property."""
        self.pop(name, None)

    return property(fget, fset, fdel, doc=doc)


refids_property = multi_string_property(
    "REFID",
    """REFID

Purpose:
    REFID acts as a key for associated iCalendar entities.

Conformance:
    Since :rfc:`9253`,
    this property can be specified zero or more times in any iCalendar component.

Description:
    The value of this property is free-form text that creates an
    identifier for associated components.
    All components that use the same REFID value are associated through
    that value and can be located or retrieved as a group.
    For example, all of the events in a travel itinerary
    would have the same REFID value, so as to be grouped together.

Examples:
    The following is an example of this property.

    .. code-block:: text

        REFID:itinerary-2014-11-17

    Use a REFID to associate several VTODOs:

    .. code-block:: pycon

        >>> from icalendar import Todo
        >>> todo_1 = Todo.new(
        ...     summary="turn off stove",
        ...     refids=["travel", "alps"]
        ... )
        >>> todo_2 = Todo.new(
        ...     summary="pack backpack",
        ...     refids=["travel", "alps"]
        ... )
        >>> todo_1.refids == todo_2.refids
        True

.. note::

    List modifications do not modify the component.
""",
)


__all__ = [
    "CONCEPTS_TYPE_SETTER",
    "LINKS_TYPE_SETTER",
    "RELATED_TO_TYPE_SETTER",
    "attendees_property",
    "busy_type_property",
    "categories_property",
    "class_property",
    "color_property",
    "comments_property",
    "concepts_property",
    "conferences_property",
    "contacts_property",
    "create_single_property",
    "description_property",
    "descriptions_property",
    "duration_property",
    "exdates_property",
    "get_duration_property",
    "get_end_property",
    "get_start_end_duration_with_validation",
    "get_start_property",
    "images_property",
    "links_property",
    "location_property",
    "multi_language_text_property",
    "multi_string_property",
    "organizer_property",
    "priority_property",
    "property_del_duration",
    "property_doc_duration_template",
    "property_get_duration",
    "property_set_duration",
    "rdates_property",
    "refids_property",
    "related_to_property",
    "rfc_7953_dtend_property",
    "rfc_7953_dtstart_property",
    "rfc_7953_duration_property",
    "rfc_7953_end_property",
    "rrules_property",
    "sequence_property",
    "set_duration_with_locking",
    "set_end_with_locking",
    "set_start_with_locking",
    "single_int_property",
    "single_utc_property",
    "source_property",
    "status_property",
    "summary_property",
    "transparency_property",
    "uid_property",
    "url_property",
]



"""This file collects errors that the OSS FUZZ build has found."""

from datetime import time

import pytest

from icalendar.cal.calendar import Calendar
from icalendar.prop import vDDDLists


def test_stack_is_empty():
    """If we get passed an invalid string, we expect to get a ValueError."""
    with pytest.raises(ValueError):
        Calendar.from_ical("END:CALENDAR")


def test_vdd_list_type_mismatch():
    """If we pass in a string type, we expect it to be converted to bytes"""
    vddd_list = vDDDLists([time(hour=6, minute=6, second=6)])
    assert vddd_list.to_ical() == b"060606"



"""This tests the parsing of the VAVAILABILITY component as defined in :rfc:`7953`."""

from datetime import date, datetime, timedelta, timezone

import pytest

from icalendar import BUSYTYPE, Availability, Available, Calendar
from icalendar.error import IncompleteComponent


@pytest.fixture(params=[Availability, Available])
def rfc_7953_component_class(request):
    """The component class."""
    return request.param


@pytest.fixture
def rfc_7953_component(rfc_7953_component_class):
    """The component class."""
    return rfc_7953_component_class()


def test_uid(availabilities):
    """Test the UID property."""
    availability: Availability = availabilities.rfc_7953_1
    assert availability.uid == "0428C7D2-688E-4D2E-AC52-CD112E2469DF"


def test_organizer(availabilities):
    """Test the ORGANIZER property."""
    availability: Availability = availabilities.rfc_7953_1
    assert availability.organizer == "mailto:bernard@example.com"
    assert availability.organizer.email == "bernard@example.com"


def test_dtstamp(availabilities, tzp):
    """Test the DTSTAMP property."""
    availability: Availability = availabilities.rfc_7953_1
    assert availability.DTSTAMP == tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))


def test_subcomponents(availabilities):
    """Test the subcomponents."""
    availability: Availability = availabilities.rfc_7953_1
    assert len(availability.subcomponents) == 1
    assert availability.available == availability.subcomponents


def test_busy_type_default():
    assert Availability().busy_type == BUSYTYPE.BUSY_UNAVAILABLE


def test_set_busy_type():
    availability = Availability()
    availability.busy_type = BUSYTYPE.BUSY
    assert availability.busy_type == BUSYTYPE.BUSY
    assert availability["BUSYTYPE"] == "BUSY"


def test_new_sets_default(test_uid):
    availability = Availability.new()
    assert availability.busy_type == BUSYTYPE.BUSY_UNAVAILABLE
    assert availability.uid == test_uid


def test_create_new_availability_with_subsomponents(dont_validate_new):
    avaiable = Available.new(summary="Available1")
    availability = Availability.new(busy_type=BUSYTYPE.BUSY, components=[avaiable])
    assert len(availability.available) == 1
    assert availability.available[0].summary == "Available1"


def test_avaiable_needs_start():
    """Start is required and not set automatically"""
    with pytest.raises(ValueError) as e:
        Available.new()
    assert "Available start must be a datetime with a timezone" in str(e.value)


@pytest.mark.parametrize("key", ["start", "end"])
@pytest.mark.parametrize(
    "value", [datetime(2011, 10, 5, 13, 32, 25), date(2011, 10, 5)]
)
def test_avaiable_needs_timezone(key, value, rfc_7953_component_class):
    """Start is required and not set automatically"""
    kw = {key: value}
    with pytest.raises(ValueError) as e:
        rfc_7953_component_class.new(**kw)
    assert f"{key} must be a datetime with a timezone" in str(e.value)


def test_create_with_start(tzp):
    """Create with a valid start."""
    start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    available = Available.new(start=start)
    assert available.start == start


def test_create_with_start_and_end(tzp):
    """Create with a valid start and end."""
    start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    end = tzp.localize_utc(datetime(2011, 10, 5, 14, 32, 25))
    available = Available.new(start=start, end=end)
    assert available.start == start
    assert available.end == end


def test_duration_from_end(rfc_7953_component, tzp):
    """Check the computation of end and duration."""
    rfc_7953_component.start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    rfc_7953_component.end = tzp.localize_utc(datetime(2011, 10, 5, 14, 32, 25))
    assert rfc_7953_component.duration == timedelta(hours=1)


def test_duration_and_end_from_duration(rfc_7953_component, tzp):
    """Check the computation of end and duration."""
    rfc_7953_component.start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    rfc_7953_component.DURATION = timedelta(hours=1)
    assert rfc_7953_component.duration == timedelta(hours=1)
    assert rfc_7953_component.end == tzp.localize_utc(datetime(2011, 10, 5, 14, 32, 25))


def test_start_only_has_no_duration(rfc_7953_component, tzp):
    """Check the computation of end and duration."""
    rfc_7953_component.start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    assert rfc_7953_component.duration is None
    assert rfc_7953_component.end is None


def test_missing_info_for_end(rfc_7953_component):
    """We miss information and cannot compute the end."""
    rfc_7953_component.DURATION = timedelta(hours=1)
    with pytest.raises(IncompleteComponent) as e:
        rfc_7953_component.end  # noqa: B018
    assert "Cannot compute end without start" in str(e.value)


def test_missing_info_for_duration(rfc_7953_component, tzp):
    """We only have a duration."""
    rfc_7953_component.end = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    with pytest.raises(IncompleteComponent) as e:
        rfc_7953_component.duration  # noqa: B018
    assert "Cannot compute duration without start" in str(e.value)


def test_example_2(availabilities, tzp):
    """Test that we get all the right values after parsing."""
    a: Availability = availabilities.rfc_7953_2
    assert a.organizer == "mailto:bernard@example.com"
    assert a.uid == "84D0F948-7FC6-4C1D-BBF3-BA9827B424B5"
    assert a.stamp == datetime(2011, 10, 5, 13, 32, 25, tzinfo=timezone.utc)
    assert a.start == tzp.localize(datetime(2011, 10, 2), "America/Montreal")
    assert a.end == tzp.localize(datetime(2011, 12, 2), "America/Montreal")
    a1 = a.available[0]
    assert a1.uid == "7B33093A-7F98-4EED-B381-A5652530F04D"
    assert a1.summary == "Monday to Thursday from 9:00 to 17:00"
    assert a1.start == tzp.localize(datetime(2011, 10, 2, 9), "America/Montreal")
    assert a1.end == tzp.localize(datetime(2011, 10, 2, 17), "America/Montreal")
    assert len(a1.rrules) == 1
    assert a1.rrules[0]["freq"] == ["WEEKLY"]
    assert a1.rrules[0]["byday"] == ["MO", "TU", "WE", "TH"]
    assert a1.location == "Main Office"
    assert a1.exdates == []
    a2 = a.available[1]
    assert a2.uid == "DF39DC9E-D8C3-492F-9101-0434E8FC1896"
    assert a2.summary == "Friday from 9:00 to 12:00"
    assert a2.start == tzp.localize(datetime(2011, 10, 6, 9), "America/Montreal")
    assert a2.end == tzp.localize(datetime(2011, 10, 6, 12), "America/Montreal")
    assert len(a2.rrules) == 1
    assert a2.rrules[0]["freq"] == ["WEEKLY"]
    assert a2.location == "Branch Office"
    assert a2.exdates == []


def test_example_3(calendars, tzp):
    """Test that we get all the right values after parsing."""
    calendar: Calendar = calendars.rfc_7953_3
    a: Availability = calendar.availabilities[0]
    assert a.organizer == "mailto:bernard@example.com"
    assert a.uid == "BE082249-7BDD-4FE0-BDBA-DE6598C32FC9"
    assert a.stamp == datetime(2011, 10, 5, 13, 32, 25, tzinfo=timezone.utc)
    assert a.start == tzp.localize(datetime(2011, 10, 2), "America/Montreal")
    assert a.end == tzp.localize(datetime(2011, 10, 23, 3), "America/Montreal")
    a1 = a.available[0]
    assert a1.uid == "54602321-CEDB-4620-9099-757583263981"
    assert a1.summary == "Monday to Friday from 9:00 to 17:00"
    assert a1.start == tzp.localize(datetime(2011, 10, 2, 9), "America/Montreal")
    assert a1.end == tzp.localize(datetime(2011, 10, 2, 17), "America/Montreal")
    assert len(a1.rrules) == 1
    assert a1.rrules[0]["freq"] == ["WEEKLY"]
    assert a1.rrules[0]["byday"] == ["MO", "TU", "WE", "TH", "FR"]
    assert a1.location == "Montreal"
    assert a1.exdates == []
    assert len(calendar.availabilities) == 3



"""Issue #350 - Ignore X-... properties also at end of file?

https://github.com/collective/icalendar/issues/350
"""


def test_issue_350(calendars):
    calendar = list(calendars.issue_350.walk("X-COMMENT"))
    assert len(calendar) == 0, "X-COMMENT at the end of the file was parsed"



"""Issue #168 - Parsing invalid icalendars fails without any warning

https://github.com/collective/icalendar/issues/168
"""


def test_issue_168_parsing_inavlid_calendars_no_warning(calendars):
    expected_error = (
        None,
        "Content line could not be parsed into parts: 'X-APPLE-RADIUS=49.91307046514149': X-APPLE-RADIUS=49.91307046514149",
    )
    assert expected_error in calendars.issue_168_input.walk("VEVENT")[0].errors
    assert (
        calendars.issue_168_input.to_ical()
        == calendars.issue_168_expected_output.raw_ics
    )



"""Tests for icalendar relationships as per RFC 9253."""

import pytest

from icalendar import (
    Alarm,
    Availability,
    Available,
    Calendar,
    Event,
    FreeBusy,
    Journal,
    Timezone,
    TimezoneDaylight,
    TimezoneStandard,
    Todo,
)
from icalendar.prop import vUri


@pytest.fixture(
    params=(
        Alarm,
        Availability,
        Available,
        Calendar,
        Event,
        FreeBusy,
        Journal,
        Timezone,
        TimezoneDaylight,
        TimezoneStandard,
        Todo,
    )
)
def link_component_class(request):
    """The classes to test.

    rfc:`9253`::

        Conformance: Link: This property can be specified zero or
        more times in any iCalendar component.

    """
    return request.param


@pytest.fixture
def link_component(link_component_class):
    """An instance of the component class to test."""
    return link_component_class()


def test_no_links_by_default(link_component):
    """By default, components have no LINK properties."""
    assert "LINK" not in link_component
    assert link_component.links == []


def test_strings_are_converted_to_vUri(link_component):
    """We set the URI."""
    link_component.links = ["https://example.com"]
    assert link_component.links == [vUri("https://example.com")]
    assert isinstance(link_component.links[0], vUri)
    assert link_component.links[0].VALUE == "URI"



"""Test the CONCEPT property."""

import pytest

from icalendar import Component, vUri


@pytest.fixture
def component():
    """Return a basic component for testing."""
    return Component()


def test_concept_is_converted_to_a_vUri(component):
    """We want a vUri list!"""
    component.concepts = ["file://a/concept"]
    assert len(component.concepts) == 1
    assert component.concepts[0].uri == "file://a/concept"
    assert isinstance(component.concepts[0], vUri)


def test_new_component_with_concept():
    """We can use new with just a string."""
    component = Component.new(concepts="https://gihub.com/collective/icalendar/issues")
    assert len(component.concepts) == 1
    assert component.concepts[0].uri == "https://gihub.com/collective/icalendar/issues"
    assert isinstance(component.concepts[0], vUri)


def test_delete_concepts(component: Component):
    del component.concepts
    component.concepts = ["ftp://asd"]
    del component.concepts
    assert component.concepts == []


def test_delete_concepts_with_none(component: Component):
    component.concepts = ["ftp://asd"]
    component.concepts = None
    assert component.concepts == []



"""This module contains compatibility code for older Python versions."""


try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        Self = "Self"

try:
    import zoneinfo
except ImportError:
    import backports.zoneinfo as zoneinfo

ZoneInfo = zoneinfo.ZoneInfo

__all__ = [
    "Self",
    "ZoneInfo",
    "zoneinfo",
]



"""This tests the creation of new components.

To aid creating valid calendars with a few lines of code, the components
receive a new() classmethod.
This method MUST use property setters to set the properties defined in the keywords.

New keywords can be added over time.

See https://github.com/collective/icalendar/issues/843
"""

from __future__ import annotations

import itertools
import traceback
from datetime import date, datetime, timedelta, timezone
from typing import Any, Callable, Optional

import pytest

from icalendar import (
    Alarm,
    Availability,
    Available,
    Calendar,
    Component,
    Event,
    FreeBusy,
    Journal,
    Todo,
    vCalAddress,
)
from icalendar.compatibility import ZoneInfo
from icalendar.enums import BUSYTYPE
from icalendar.prop import vText, vUid, vUri, vXmlReference

from .conftest import NOW_UTC, UID_DEFAULT

UTC = timezone.utc

# Test parametrization

param_summary_components = pytest.mark.parametrize(
    "component", [Event, Todo, Alarm, Journal, Available, Availability]
)
param_description_components = pytest.mark.parametrize(
    "component", [Event, Todo, Alarm, Journal, Available, Availability]
)
COMPONENTS_DTSTAMP = {
    Component,
    Event,
    Journal,
    Todo,
    FreeBusy,
    Available,
    Availability,
}
COMPONENTS_DTSTAMP_AUTOMATIC = {Event, Journal, Todo, FreeBusy, Available, Availability}
COMPONENTS_UID_AUTOMATIC = {Event, Todo, Journal, Available, Availability}
COMPONENTS_UID = {Event, Todo, Journal, Alarm, Calendar, Available, Availability}
COMPONENTS_SEQUENCE = {Event, Todo, Journal, Availability}
COMPONENTS_CATEGORIES = {Event, Journal, Todo, Calendar, Availability, Available}
COMPONENTS_ORGANIZER = {Availability, Event, FreeBusy, Journal, Todo}
COMPONENTS_LOCATION = {Availability, Available, Event, Todo}
COMPONENTS_URL = {Availability, Event, Todo, Journal, FreeBusy, Calendar}
COMPONENTS_BUSYTYPE = {Availability}
COMPONENTS_DESCRIPTION = {
    Event,
    Todo,
    Journal,
    Alarm,
    Available,
    Availability,
    Calendar,
}
COMPONENTS_SUMMARY = {Event, Todo, Journal, Alarm, Available, Availability}
COMPONENTS_COMMENT = {
    Event,
    Todo,
    Journal,
    FreeBusy,
    Available,
    Availability,
}  # Standard and Daylight
COMPONENTS_PRIORITY = {Event, Todo, Availability}
COMPONENTS_CONTACT = {Event, Todo, Journal, FreeBusy, Available, Availability}
COMPONENTS_START_END = {Event, Todo, FreeBusy, Available, Availability}
COMPONENTS_STATUS = {Event, Todo, Journal}
COMPONENTS_ATTENDEES = {Event, Todo, Journal, Alarm}
# RFC 9253 properties are defines on ALL
# So, if you add new components, do not forget to add them here.
COMPONENTS_LINKS = COMPONENTS_RELATED_TO = COMPONENTS_CONCEPTS = COMPONENTS_REFID = {
    Alarm,
    Availability,
    Available,
    Calendar,
    Event,
    FreeBusy,
    Journal,
    Todo,
}


@param_summary_components
def test_summary_default(component):
    """Test the summary property default."""
    c = component()
    assert c.summary is None


@param_description_components
def test_description_default(component):
    """Test the summary property default."""
    c = component()
    assert c.description is None


@param_summary_components
def test_summary_delete(component):
    """Test the summary property default."""
    c = component()
    c.summary = "alksdj"
    del c.summary
    assert c.summary is None
    del c.summary
    assert c.summary is None


@param_description_components
def test_description_delete(component):
    """Test the summary property default."""
    c = component()
    c.description = "alksdj"
    del c.description
    assert c.description is None
    del c.description
    assert c.description is None


param_summary = pytest.mark.parametrize(
    "summary", ["akshdkjahskjdhas", "This is a more\ncomplex summary.", None]
)
param_description = pytest.mark.parametrize(
    "description", ["akshdkjahskjdhas", "This is a more\ncomplex summary.", None]
)


def assert_summary_equals(component: Component, summary: Optional[str]):
    """Check this is the summary."""
    assert_property_equals(component, "summary", summary)


def assert_description_equals(component: Component, summary: Optional[str]):
    """Check this is the summary."""
    assert_property_equals(component, "description", summary)


def assert_property_equals(component: Component, name: str, text: Optional[str]):
    """Check the property."""
    assert getattr(component, name.lower()) == text
    if text is None:
        assert name.upper() not in component
    else:
        assert component[name.upper()] == text


@param_summary_components
@param_summary
def test_set_summary(component, summary):
    """Test the summary property default."""
    c = component()
    c.summary = summary
    assert_summary_equals(c, summary)


@param_summary_components
@param_summary
def test_new_with_summary(component, summary, dont_validate_new):
    """Test the summary property default."""
    assert_summary_equals(component.new(summary=summary), summary)


@param_description_components
@param_description
def test_set_description(component, description):
    """Test the description property default."""
    c = component()
    c.description = description
    assert_description_equals(c, description)


@param_description_components
@param_description
def test_new_with_description(component, description, dont_validate_new):
    """Test the description property default."""
    assert_description_equals(component.new(description=description), description)


def new_journal_description(description) -> Journal:
    """use new()"""
    return Journal.new(description=description)


def set_journal_description(description) -> Journal:
    """Set the description"""
    journal = Journal()
    journal.descriptions = description
    return journal


@pytest.mark.parametrize(
    ("description", "expected_description"),
    [
        (None, []),
        ([], []),
        ("one description", ["one description"]),
        (("desc12", "desc23"), ["desc12", "desc23"]),
    ],
)
@pytest.mark.parametrize(
    "get_journal", [new_journal_description, set_journal_description]
)
def test_journal_description_is_a_list(
    get_journal: Callable[..., Journal], description, expected_description
):
    """A jounal entry can have several descriptions."""
    journal = get_journal(description)
    assert journal.descriptions == expected_description
    if not description:
        assert "DESCRIPTION" not in journal
    else:
        assert "DESCRIPTION" in journal


def test_multiple_descriptions_are_concatenated():
    """For compatibility we also provide the description method that concatenates descriptions."""
    journal = Journal.new(description=("one description", "two descriptions"))
    assert journal.description == "one description\r\n\r\ntwo descriptions"


def component_setter(
    component_class: type[Component], property_name: str, value: Any
) -> Component:
    """We set the value of the component's attribute."""
    component = component_class()
    try:
        setattr(component, property_name, value)
    except TypeError:
        # we show the exception in case the test fails for easier debug
        traceback.print_exc()
        pytest.skip("Cannot set value")
    return component


def component_with_new(
    component_class: type[Component], property_name: str, value: Any
) -> Component:
    """Use new() to create the component."""
    kw = {property_name.lower(): value}
    return component_class.new(**kw)


def assert_component_attribute_has_value(
    component: Component, property_name: str, expected_value: Any, message
):
    """Make  sure that a component's attribute has a value."""
    actual_value = getattr(component, property_name)
    assert actual_value == expected_value, (
        f"The attribute {property_name} of {component} "
        f"should be {expected_value!r} but got {actual_value!r}.\n{message}"
    )


automatic_time_test_cases = list(
    itertools.chain.from_iterable(
        [
            (
                COMPONENTS_DTSTAMP.copy(),
                property_name,
                key,
                date(2023, 10, 21),
                datetime(2023, 10, 21, tzinfo=timezone.utc),
                True,
                f"{key} becomes a UTC value 1",
            ),
            (
                COMPONENTS_DTSTAMP.copy(),
                property_name,
                key,
                datetime(2023, 10, 22),
                datetime(2023, 10, 22, tzinfo=timezone.utc),
                True,
                f"{key} becomes a UTC value 2",
            ),
            (
                COMPONENTS_DTSTAMP.copy(),
                property_name,
                key,
                datetime(2023, 10, 23, 12, 30, tzinfo=timezone.utc),
                datetime(2023, 10, 23, 12, 30, tzinfo=ZoneInfo("UTC")),
                True,
                f"{key} becomes a UTC value 3",
            ),
            (
                COMPONENTS_DTSTAMP.copy(),
                property_name,
                key,
                datetime(2023, 10, 24, 21, 0, 1, tzinfo=timezone(timedelta(hours=1))),
                datetime(2023, 10, 24, 20, 0, 1, tzinfo=ZoneInfo("UTC")),
                True,
                f"{key} becomes a UTC value 4",
            ),
            (
                COMPONENTS_DTSTAMP_AUTOMATIC.copy(),
                property_name,
                key,
                None,
                NOW_UTC,
                key == "DTSTAMP",
                f"we use the current time to create a datetime for {key}",
            ),
            (
                COMPONENTS_DTSTAMP - COMPONENTS_DTSTAMP_AUTOMATIC,
                property_name,
                key,
                None,
                None,
                False,
                f"{key} is not automatically set",
            ),
        ]
        for property_name, key in (
            ("stamp", "DTSTAMP"),
            ("created", "CREATED"),
            ("last_modified", "LAST-MODIFIED"),
        )
    )
)
for (
    component_classes,
    property_name,
    _key,
    _value,
    _expected_value,
    _expected,
    _message,
) in automatic_time_test_cases:
    if property_name != "stamp" and FreeBusy in component_classes:
        # FreeBusy only has stamp.
        component_classes.remove(FreeBusy)


attendee1 = vCalAddress("mailto:attendee1@example.com")
attendee2 = vCalAddress("mailto:attendee2@test.test")


new_test_cases = [
    (
        COMPONENTS_UID,
        "uid",
        "uid",
        "test UID",
        "test UID",
        True,
        "Set the UID property",
    ),
    (
        COMPONENTS_UID_AUTOMATIC,
        "uid",
        "uid",
        None,
        UID_DEFAULT,
        True,
        "Set the UID property by default for these components",
    ),
    (
        COMPONENTS_UID - COMPONENTS_UID_AUTOMATIC,
        "uid",
        "uid",
        None,
        "",
        False,
        "UID is not automatically set",
    ),
    (
        {Event, Todo, Journal, Availability, Available, FreeBusy},
        "start",
        "dtstart",
        datetime(2023, 10, 24, 21, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        datetime(2023, 10, 24, 21, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        True,
        "set the start",
    ),
    (
        {Event, Available, FreeBusy},
        "end",
        "dtend",
        datetime(2023, 10, 24, 22, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        datetime(2023, 10, 24, 22, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        True,
        "set the end",
    ),
    (
        {Todo},
        "end",
        "due",
        datetime(2023, 10, 24, 22, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        datetime(2023, 10, 24, 22, 0, 1, tzinfo=ZoneInfo("Europe/Berlin")),
        True,
        "set the end",
    ),
    (
        {Todo, Event, Calendar, Journal},
        "color",
        "color",
        "red",
        "red",
        True,
        "set the color",
    ),
    (
        COMPONENTS_SEQUENCE,
        "sequence",
        "SEQUENCE",
        1,
        1,
        True,
        "set the sequence",
    ),
    (
        COMPONENTS_SEQUENCE,
        "sequence",
        "SEQUENCE",
        None,
        0,
        False,
        "get the default the sequence",
    ),
    (
        COMPONENTS_CATEGORIES,
        "categories",
        "CATEGORIES",
        ["cat1", "cat2"],
        ["cat1", "cat2"],
        True,
        "set the categories",
    ),
    (
        COMPONENTS_CATEGORIES,
        "categories",
        "CATEGORIES",
        (),
        [],
        False,
        "categories are absent",
    ),
    (
        COMPONENTS_ORGANIZER,
        "organizer",
        "ORGANIZER",
        "mailto:bernard@example.com",
        "mailto:bernard@example.com",
        True,
        "set the organizer",
    ),
    (
        COMPONENTS_ORGANIZER,
        "organizer",
        "ORGANIZER",
        None,
        None,
        False,
        "no organizer by default",
    ),
    (
        COMPONENTS_LOCATION,
        "location",
        "LOCATION",
        "Berlin",
        "Berlin",
        True,
        "set the location",
    ),
    (
        COMPONENTS_LOCATION,
        "location",
        "LOCATION",
        None,
        None,
        False,
        "no location by default",
    ),
    (
        COMPONENTS_URL,
        "url",
        "URL",
        "https://icalendar.readthedocs.io/",
        "https://icalendar.readthedocs.io/",
        True,
        "set the url",
    ),
    (
        COMPONENTS_URL,
        "url",
        "URL",
        None,
        "",
        False,
        "no url by default",
    ),
    (
        COMPONENTS_BUSYTYPE,
        "busy_type",
        "BUSYTYPE",
        BUSYTYPE.BUSY,
        BUSYTYPE.BUSY,
        True,
        "set the url",
    ),
    (
        COMPONENTS_BUSYTYPE,
        "busy_type",
        "BUSYTYPE",
        None,
        BUSYTYPE.BUSY_UNAVAILABLE,
        False,
        "default busy type",
    ),
    (
        COMPONENTS_DESCRIPTION,
        "description",
        "DESCRIPTION",
        "This describes the component a bit more in detail.",
        "This describes the component a bit more in detail.",
        True,
        "set the description",
    ),
    (
        COMPONENTS_DESCRIPTION,
        "description",
        "DESCRIPTION",
        None,
        None,
        False,
        "no description by default",
    ),
    (
        COMPONENTS_SUMMARY,
        "summary",
        "SUMMARY",
        "component summary",
        "component summary",
        True,
        "set the summary",
    ),
    (
        COMPONENTS_SUMMARY,
        "summary",
        "SUMMARY",
        None,
        None,
        False,
        "no summary by default",
    ),
    (
        COMPONENTS_PRIORITY,
        "priority",
        "PRIORITY",
        8,
        8,
        True,
        "priority is set",
    ),
    (
        COMPONENTS_PRIORITY,
        "priority",
        "PRIORITY",
        None,
        0,
        False,
        "no priority by default",
    ),
    (
        COMPONENTS_COMMENT,
        "comments",
        "COMMENT",
        None,
        [],
        False,
        "no comments",
    ),
    (
        COMPONENTS_COMMENT,
        "comments",
        "COMMENT",
        "This is a comment",
        ["This is a comment"],
        True,
        "one comment is set",
    ),
    (
        COMPONENTS_COMMENT,
        "comments",
        "COMMENT",
        ["one", "two", "three"],
        ["one", "two", "three"],
        True,
        "several comments",
    ),
    (
        COMPONENTS_CONTACT,
        "contacts",
        "CONTACT",
        None,
        [],
        False,
        "no contacts",
    ),
    (
        COMPONENTS_CONTACT,
        "contacts",
        "CONTACT",
        "This is a comment",
        ["This is a comment"],
        True,
        "one contact is set",
    ),
    (
        COMPONENTS_CONTACT,
        "contacts",
        "CONTACT",
        ["one", "two", "three"],
        ["one", "two", "three"],
        True,
        "several contacts",
    ),
    (
        {Event},
        "transparency",
        "TRANSP",
        "TRANSPARENT",
        "TRANSPARENT",
        True,
        "value set to TRANSPARENT",
    ),
    (
        {Event},
        "transparency",
        "TRANSP",
        "OPAQUE",
        "OPAQUE",
        True,
        "value set to OPAQUE",
    ),
    (
        {Event},
        "transparency",
        "TRANSP",
        None,
        "OPAQUE",
        False,
        "value is not set, defaults to OPAQUE",
    ),
    (
        COMPONENTS_STATUS,
        "status",
        "STATUS",
        "CANCELLED",
        "CANCELLED",
        True,
        "value set to CANCELLED",
    ),
    (
        COMPONENTS_STATUS,
        "status",
        "STATUS",
        "NEEDS-ACTION",
        "NEEDS-ACTION",
        True,
        "value set to NEEDS-ACTION",
    ),
    (
        COMPONENTS_STATUS,
        "status",
        "STATUS",
        None,
        "",
        False,
        "value is not set",
    ),
    (
        COMPONENTS_STATUS,
        "status",
        "STATUS",
        "",
        "",
        False,
        "value is not set",
    ),
    (
        COMPONENTS_ATTENDEES,
        "attendees",
        "ATTENDEE",
        [],
        [],
        True,
        "value is not set",
    ),
    (
        COMPONENTS_ATTENDEES,
        "attendees",
        "ATTENDEE",
        None,
        [],
        False,
        "value is empty",
    ),
    (
        COMPONENTS_ATTENDEES,
        "attendees",
        "ATTENDEE",
        [attendee1],
        [attendee1],
        True,
        "one attendee",
    ),
    (
        COMPONENTS_ATTENDEES,
        "attendees",
        "ATTENDEE",
        attendee2,
        [attendee2],
        True,
        "one attendee",
    ),
    (
        COMPONENTS_ATTENDEES,
        "attendees",
        "ATTENDEE",
        [attendee1, attendee2],
        [attendee1, attendee2],
        True,
        "two attendees",
    ),
]

rfc_7986_test_cases = [
    (
        {Calendar},
        "last_modified",
        "LAST-MODIFIED",
        datetime(2011, 10, 5, 13, 32, 25),
        datetime(2011, 10, 5, 13, 32, 25, tzinfo=UTC),
        True,
        "value set to 2011-10-05T13:32:25 UTC",
    ),
    (
        {Calendar},
        "last_modified",
        "LAST-MODIFIED",
        None,
        None,
        False,
        "value is not set",
    ),
    (
        {Calendar},
        "source",
        "SOURCE",
        "https://github.com/collective/icalendar",
        "https://github.com/collective/icalendar",
        True,
        "setting a string",
    ),
    (
        {Calendar},
        "refresh_interval",
        "REFRESH-INTERVAL",
        timedelta(hours=1),
        timedelta(hours=1),
        True,
        "setting a timedelta",
    ),
    (
        {Calendar},
        "refresh_interval",
        "REFRESH-INTERVAL",
        None,
        None,
        False,
        "setting nothing",
    ),
]


rfc_9253_link_values = [
    vUri("https://123"),
    vUid("123-123-123"),
    vXmlReference("http://example.com"),
]

rfc_9253_related_to_values = [
    vUri("https://123"),
    vUid("123-123-123"),
    vText("nananana", params={"RELTYPE": "SIBLING"}),
]
rfc_9253_test_cases = [
    (
        COMPONENTS_LINKS,
        "links",
        "LINK",
        None,
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_LINKS,
        "links",
        "LINK",
        [],
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_LINKS,
        "links",
        "LINK",
        ["https://123"],
        [vUri("https://123")],
        True,
        "set one value",
    ),
    (
        COMPONENTS_LINKS,
        "links",
        "LINK",
        rfc_9253_link_values,
        rfc_9253_link_values,
        True,
        "set several values",
    ),
    (
        COMPONENTS_RELATED_TO,
        "related_to",
        "RELATED-TO",
        None,
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_RELATED_TO,
        "related_to",
        "RELATED-TO",
        [],
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_RELATED_TO,
        "related_to",
        "RELATED-TO",
        ["https://123"],
        [vText("https://123")],
        True,
        "set one value",
    ),
    (
        COMPONENTS_RELATED_TO,
        "related_to",
        "RELATED-TO",
        rfc_9253_related_to_values,
        rfc_9253_related_to_values,
        True,
        "set several values",
    ),
    (
        COMPONENTS_CONCEPTS,
        "concepts",
        "CONCEPT",
        None,
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_CONCEPTS,
        "concepts",
        "CONCEPT",
        [],
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_CONCEPTS,
        "concepts",
        "CONCEPT",
        ["https://123", vUri("https://asd")],
        [vUri("https://123"), vUri("https://asd")],
        True,
        "set two values",
    ),
    (
        COMPONENTS_REFID,
        "refids",
        "REFID",
        None,
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_REFID,
        "refids",
        "REFID",
        [],
        [],
        False,
        "setting nothing",
    ),
    (
        COMPONENTS_REFID,
        "refids",
        "REFID",
        "itinerary-2014-11-17",
        ["itinerary-2014-11-17"],
        True,
        "set a value",
    ),
    (
        COMPONENTS_REFID,
        "refids",
        "REFID",
        ["itinerary-2014-11-17", "itinerary-2014-11-17-2"],
        ["itinerary-2014-11-17", "itinerary-2014-11-17-2"],
        True,
        "set two values",
    ),
]


@pytest.mark.parametrize(
    (
        "component_classes",
        "property_name",
        "key",
        "initial_value",
        "expected_value",
        "key_present",
        "message",
    ),
    automatic_time_test_cases
    + new_test_cases
    + rfc_7986_test_cases
    + rfc_9253_test_cases,
)
@pytest.mark.parametrize(
    "create_component_with_property", [component_setter, component_with_new]
)
def test_properties_and_new(
    create_component_with_property,
    component_classes,
    property_name,
    initial_value,
    expected_value,
    key_present,
    key,
    message,
    dont_validate_new,
):
    """We set and get the dtstamp."""
    for component_class in component_classes:
        print(
            "processing:",
            component_class.name,
            "->",
            property_name,
            "=",
            initial_value,
            ":",
            message,
        )
        component = create_component_with_property(
            component_class, property_name, initial_value
        )
        if (
            create_component_with_property is component_with_new
            or initial_value is not None
        ):
            assert (key in component) == key_present, message
            # the setter does not create default values
            # so we only check it if present
            assert_component_attribute_has_value(
                component, property_name, expected_value, message
            )


@pytest.mark.parametrize("component_class", COMPONENTS_START_END)
def test_end_must_be_after_start(tzp, component_class):
    """The end must be after the start."""
    with pytest.raises(ValueError) as e:
        component_class.new(
            start=tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25)),
            end=tzp.localize_utc(datetime(2011, 10, 5, 12, 32, 25)),
        )
    assert "end must be after start" in str(e.value)


@pytest.mark.parametrize("component_class", COMPONENTS_START_END)
def test_start_and_end_can_be_the_same(tzp, component_class):
    """The end must be after the start."""
    start = tzp.localize_utc(datetime(2011, 10, 5, 13, 32, 25))
    c = component_class.new(start=start, end=start)
    assert c.start == start
    assert c.end == start
    assert c.duration == timedelta(0)


def test_journal_start():
    """Journal does not have an end."""
    j = Journal.new(start=datetime(2011, 10, 5, 13, 32, 25))
    assert j.start == datetime(2011, 10, 5, 13, 32, 25)


def test_modify_attendees():
    """We modify the list and see of that works."""
    attendees = [attendee1]
    j = Journal.new(attendees=attendees)
    j.attendees.append(attendee2)
    assert j.attendees == [attendee1, attendee2]
    ics = j.to_ical().decode()
    assert str(attendee1) in ics
    assert str(attendee2) in ics


def test_modify_default_attendees():
    """Empty attendees need to be modifiable."""
    todo = Todo.new()
    todo.attendees.append(attendee1)
    assert todo.attendees == [attendee1]
    ics = todo.to_ical().decode()
    assert str(attendee1) in ics


def test_empty_attendees_to_ical():
    """We should be able to have no attendees set."""
    e = Event.new()
    e.attendees = []
    assert "ATTENDEE" not in e.to_ical().decode()


def test_modify_empty_attendees_1():
    """Empty value set and modified."""
    e = Event()
    attendees = []
    e.attendees = attendees
    e.attendees.append(attendee1)
    assert e.attendees == [attendee1] == attendees


def test_modify_empty_attendees_2():
    """Empty value set and modified."""
    e = Event()
    attendees = []
    e.attendees = attendees
    attendees.append(attendee1)
    assert e.attendees == [attendee1] == attendees



"""This tests the compatibility with RFC 7529.

See
- https://github.com/collective/icalendar/issues/655
- https://www.rfc-editor.org/rfc/rfc7529.html
"""

import pytest

from icalendar.prop import vMonth, vRecur, vSkip


@pytest.mark.parametrize(
    ("uid", "scale"),
    [
        ("4.3.1", "CHINESE"),
        ("4.3.2", "ETHIOPIC"),
        ("4.3.3", "HEBREW"),
        ("4.3.4", "GREGORIAN"),
    ],
)
def test_rscale(calendars, uid, scale):
    """Check that the RSCALE is parsed correctly."""
    event = calendars.rfc_7529.walk(select=lambda c: c.get("UID") == uid)[0]
    print(event.errors)
    rrule = event["RRULE"]
    print(rrule)
    assert rrule["RSCALE"] == [scale]


@pytest.mark.parametrize(
    ("uid", "skip"),
    [
        ("4.3.2", None),
        ("4.3.3", ["FORWARD"]),
    ],
)
def test_rscale_with_skip(calendars, uid, skip):
    """Check that the RSCALE is parsed correctly."""
    event = calendars.rfc_7529.walk(select=lambda c: c.get("UID") == uid)[0]
    recur = event["RRULE"]
    assert recur.get("SKIP") == skip


def test_leap_month(calendars):
    """Check that we can parse the leap month."""
    event = calendars.rfc_7529.walk(select=lambda c: c.get("UID") == "4.3.3")[0]
    recur = event["RRULE"]
    assert recur["BYMONTH"][0].leap is True


@pytest.mark.parametrize(
    ("ty", "recur", "ics"),
    [
        (
            vRecur,
            vRecur(rscale="CHINESE", freq="YEARLY"),
            b"RSCALE=CHINESE;FREQ=YEARLY",
        ),
        (vRecur, vRecur(bymonth=vMonth(10)), b"BYMONTH=10"),
        (vRecur, vRecur(bymonth=vMonth("5L")), b"BYMONTH=5L"),
        (vMonth, vMonth(10), b"10"),
        (vMonth, vMonth("5L"), b"5L"),
        (vSkip, vSkip.OMIT, b"OMIT"),
        (vSkip, vSkip.BACKWARD, b"BACKWARD"),
        (vSkip, vSkip.FORWARD, b"FORWARD"),
        (vSkip, vSkip("OMIT"), b"OMIT"),
        (vSkip, vSkip("BACKWARD"), b"BACKWARD"),
        (vSkip, vSkip("FORWARD"), b"FORWARD"),
        (
            vRecur,
            vRecur(rscale="GREGORIAN", freq="YEARLY", skip="FORWARD"),
            b"RSCALE=GREGORIAN;FREQ=YEARLY;SKIP=FORWARD",
        ),
        (
            vRecur,
            vRecur(rscale="GREGORIAN", freq="YEARLY", skip=vSkip.FORWARD),
            b"RSCALE=GREGORIAN;FREQ=YEARLY;SKIP=FORWARD",
        ),
    ],
)
def test_conversion(ty, recur, ics):
    """Test string conversion."""
    assert recur.to_ical() == ics
    assert ty.from_ical(ics.decode()) == recur
    assert ty.from_ical(ics.decode()).to_ical() == ics



"""The interface for timezone implementations."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from datetime import datetime, tzinfo

    from dateutil.rrule import rrule

    from icalendar import prop


class TZProvider(ABC):
    """Interface for timezone implementations."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the implementation."""

    @abstractmethod
    def localize_utc(self, dt: datetime) -> datetime:
        """Return the datetime in UTC."""

    @abstractmethod
    def localize(self, dt: datetime, tz: tzinfo) -> datetime:
        """Localize a datetime to a timezone."""

    @abstractmethod
    def knows_timezone_id(self, id: str) -> bool:
        """Whether the timezone is already cached by the implementation."""

    @abstractmethod
    def fix_rrule_until(self, rrule: rrule, ical_rrule: prop.vRecur) -> None:
        """Make sure the until value works for the rrule generated from the ical_rrule."""

    @abstractmethod
    def create_timezone(self, name: str, transition_times, transition_info) -> tzinfo:
        """Create a pytz timezone file given information."""

    @abstractmethod
    def timezone(self, name: str) -> Optional[tzinfo]:
        """Return a timezone with a name or None if we cannot find it."""

    @abstractmethod
    def uses_pytz(self) -> bool:
        """Whether we use pytz."""

    @abstractmethod
    def uses_zoneinfo(self) -> bool:
        """Whether we use zoneinfo."""


__all__ = ["TZProvider"]



"""Test the REFID property."""

import pytest

from icalendar import Component


@pytest.fixture
def component():
    """Return a basic component for testing."""
    return Component()


def test_new_component_with_concept():
    """We can use new with just a string."""
    component = Component.new(refids="refid-1")
    assert len(component.refids) == 1
    assert component.refids[0] == "refid-1"
    assert isinstance(component.refids[0], str)


def test_delete_refids(component: Component):
    del component.refids
    component.refids = ["asd"]
    del component.refids
    assert component.refids == []


def test_delete_refids_with_none(component: Component):
    component.refids = ["123"]
    component.refids = None
    assert component.refids == []



"""Test the mappings from windows to olson tzids"""

from datetime import datetime

import pytest

from icalendar import vDatetime
from icalendar.timezone.windows_to_olson import WINDOWS_TO_OLSON


def test_windows_timezone(tzp):
    """Test that the timezone is mapped correctly to olson."""
    dt = vDatetime.from_ical("20170507T181920", "Eastern Standard Time")
    expected = tzp.localize(datetime(2017, 5, 7, 18, 19, 20), "America/New_York")
    assert dt.tzinfo == expected.tzinfo
    assert dt == expected


@pytest.mark.parametrize("olson_id", WINDOWS_TO_OLSON.values())
def test_olson_names(tzp, olson_id):
    """test if all mappings actually map to valid tzids"""
    assert tzp.timezone(olson_id) is not None



"""Testing functionality that is common to date and time property types."""

from datetime import datetime, time
from typing import TYPE_CHECKING

import pytest

from icalendar import Component, vDatetime, vDDDLists, vDDDTypes, vTime
from icalendar.compatibility import ZoneInfo
from icalendar.parser_tools import to_unicode
from icalendar.prop import VPROPERTY
from icalendar.timezone.tzid import is_utc

if TYPE_CHECKING:
    from icalendar.cal.event import Event


@pytest.fixture(
    params=[
        vDatetime(datetime(2018, 1, 1, tzinfo=ZoneInfo("UTC"))),
        vTime(time(1, 1, tzinfo=ZoneInfo("UTC"))),
    ]
)
def utc_prop(request):
    return request.param


@pytest.mark.parametrize("convert", [None, vDDDLists, vDDDTypes])
def test_setting_a_utc_datetime_value_does_not_include_a_TZID(
    utc_prop: VPROPERTY, convert
):
    """Check that the VALUE parameter is correctly determined."""
    if convert:
        utc_prop = convert(utc_prop.dt)
    assert utc_prop.params.is_utc()
    component = Component()
    component.add("X-PROP", utc_prop)
    component.add("DTSTAMP", utc_prop)
    component.add("DTSTART", utc_prop)
    component.add("DUE", utc_prop)

    ical = component.to_ical().decode()
    print(ical)
    assert "TZID" not in ical


def test_converting_to_utc_puts_a_z_in_the_end(utc_prop: VPROPERTY):
    """Test that the Z is appended.

    :rfc:`5545`, TIME:
        UTC time, or absolute time, is identified by a LATIN CAPITAL
        LETTER Z suffix character, the UTC designator, appended to the
        time value.  For example, the following represents 07:00 AM UTC:

        .. code-block:: text

            070000Z

        The "TZID" property parameter MUST NOT be applied to TIME
        properties whose time values are specified in UTC.

    :rfc:`5545`, DATE-TIME:
        The date with UTC time, or absolute time, is identified by a LATIN
        CAPITAL LETTER Z suffix character, the UTC designator, appended to
        the time value.  For example, the following represents January 19,
        1998, at 0700 UTC:

        .. code-block:: text

            19980119T070000Z

        The "TZID" property parameter MUST NOT be applied to DATE-TIME
        properties whose time values are specified in UTC.

    """
    ical = to_unicode(utc_prop.to_ical())
    assert ical.endswith("Z")


def test_reformed_UTC_serialization(events):
    """Check that UTC values are still propertly serialized."""
    event: Event = events.issue_156_RDATE_with_PERIOD
    utc_value = event["DTSTART"]
    assert is_utc(utc_value.dt)
    # TODO: the UTC parameter is special - it can be present but should not be serialized
    # How should we treat the parameter?
    # I think: have it there, but do not serialize it



import re

import pytest

from icalendar import Parameters, vCalAddress
from icalendar.cal.calendar import Calendar
from icalendar.cal.event import Event


@pytest.mark.parametrize(
    ("parameter", "expected"),
    [
        # Simple parameter:value pair
        (Parameters(parameter1="Value1"), b"PARAMETER1=Value1"),
        # Parameter with list of values must be separated by comma
        (Parameters({"parameter1": ["Value1", "Value2"]}), b"PARAMETER1=Value1,Value2"),
        # Multiple parameters must be separated by a semicolon
        (
            Parameters({"RSVP": "TRUE", "ROLE": "REQ-PARTICIPANT"}),
            b"ROLE=REQ-PARTICIPANT;RSVP=TRUE",
        ),
        # Parameter values containing ',;:' must be double quoted
        (Parameters({"ALTREP": "http://www.wiz.org"}), b'ALTREP="http://www.wiz.org"'),
        # list items must be quoted separately
        (
            Parameters(
                {"MEMBER": ["MAILTO:projectA@host.com", "MAILTO:projectB@host.com"]}
            ),
            b'MEMBER="MAILTO:projectA@host.com","MAILTO:projectB@host.com"',
        ),
        (
            Parameters(
                {
                    "parameter1": "Value1",
                    "parameter2": ["Value2", "Value3"],
                    "ALTREP": ["http://www.wiz.org", "value4"],
                }
            ),
            b'ALTREP="http://www.wiz.org","value4";PARAMETER1=Value1;PARAMETER2=Value2,Value3',
        ),
        # Including empty strings
        (Parameters({"PARAM": ""}), b"PARAM="),
        # We can also parse parameter strings
        (
            Parameters(
                {"MEMBER": ["MAILTO:projectA@host.com", "MAILTO:projectB@host.com"]}
            ),
            b'MEMBER="MAILTO:projectA@host.com","MAILTO:projectB@host.com"',
        ),
        # We can also parse parameter strings
        (
            Parameters(
                {
                    "PARAMETER1": "Value1",
                    "ALTREP": ["http://www.wiz.org", "value4"],
                    "PARAMETER2": ["Value2", "Value3"],
                }
            ),
            b'ALTREP="http://www.wiz.org","value4";PARAMETER1=Value1;PARAMETER2=Value2,Value3',
        ),
    ],
)
def test_parameter_to_ical_is_inverse_of_from_ical(parameter, expected):
    assert parameter.to_ical() == expected
    assert Parameters.from_ical(expected.decode("utf-8")) == parameter


def test_parse_parameter_string_without_quotes():
    assert Parameters.from_ical("PARAM1=Value 1;PARA2=Value 2") == Parameters(
        {"PARAM1": "Value 1", "PARA2": "Value 2"}
    )


def test_parametr_is_case_insensitive():
    parameter = Parameters(parameter1="Value1")
    assert parameter["parameter1"] == parameter["PARAMETER1"] == parameter["PaRaMeTer1"]


def test_parameter_keys_are_uppercase():
    parameter = Parameters(parameter1="Value1")
    assert list(parameter.keys()) == ["PARAMETER1"]


@pytest.mark.parametrize(
    ("cn_param", "cn_quoted"),
    [
        # not double-quoted
        ("Aramis", "Aramis"),
        # if a space is present - enclose in double quotes
        ("Aramis Alameda", '"Aramis Alameda"'),
        # a single quote in parameter value - double quote the value
        ("Aramis d'Alameda", '"Aramis d\'Alameda"'),
        ("Арамис д'Аламеда", '"Арамис д\'Аламеда"'),
        # Before, double quote is replaced with single quote
        # Since RFC 6868, we replace this with ^'
        ('Aramis d"Alameda', '"Aramis d^\'Alameda"'),
    ],
)
def test_quoting(cn_param, cn_quoted):
    event = Event()
    attendee = vCalAddress("test@example.com")
    attendee.params["CN"] = cn_param
    event.add("ATTENDEE", attendee)
    assert f"ATTENDEE;CN={cn_quoted}:test@example.com" in event.to_ical().decode(
        "utf-8"
    )


def test_property_params():
    """Property parameters with values containing a COLON character, a
    SEMICOLON character or a COMMA character MUST be placed in quoted
    text."""
    cal_address = vCalAddress("mailto:john.doe@example.org")
    cal_address.params["CN"] = "Doe, John"
    ical = Calendar()
    ical.add("organizer", cal_address)

    ical_str = Calendar.to_ical(ical)
    exp_str = (
        b"""BEGIN:VCALENDAR\r\nORGANIZER;CN="Doe, John":"""
        b"""mailto:john.doe@example.org\r\nEND:VCALENDAR\r\n"""
    )

    assert ical_str == exp_str

    # other way around: ensure the property parameters can be restored from
    # an icalendar string.
    ical2 = Calendar.from_ical(ical_str)
    assert ical2.get("ORGANIZER").params.get("CN") == "Doe, John"


def test_parse_and_access_property_params(calendars):
    """Parse an ics string and access some property parameters then.
    This is a follow-up of a question received per email.

    """
    event = calendars.property_params.walk("VEVENT")[0]
    attendee = event["attendee"][0]
    assert attendee.to_ical() == b"MAILTO:rembrand@xs4all.nl"
    assert attendee.params.to_ical() == b"CN=RembrandXS;PARTSTAT=NEEDS-ACTION;RSVP=TRUE"
    assert attendee.params["cn"] == "RembrandXS"


def test_repr():
    """Test correct class representation."""
    it = Parameters(parameter1="Value1")
    assert re.match(r"Parameters\({u?'PARAMETER1': u?'Value1'}\)", str(it))



import pytest


def test_ignore_exceptions_on_broken_events_issue_104(events):
    """Issue #104 - line parsing error in a VEVENT
    (which has ignore_exceptions). Should mark the event broken
    but not raise an exception.

    https://github.com/collective/icalendar/issues/104
    """
    assert events.issue_104_mark_events_broken.errors == [
        (None, "Content line could not be parsed into parts: 'X': Invalid content line")
    ]


def test_dont_ignore_exceptions_on_broken_calendars_issue_104(calendars):
    """Issue #104 - line parsing error in a VCALENDAR
    (which doesn't have ignore_exceptions). Should raise an exception.
    """
    with pytest.raises(ValueError):
        calendars.issue_104_broken_calendar


def test_rdate_dosent_become_none_on_invalid_input_issue_464(events):
    """Issue #464 - [BUG] RDATE can become None if value is invalid
    https://github.com/collective/icalendar/issues/464
    """
    # After VALUE parameter fix, the assertion checks for the full value string
    # in the error message, making the test more flexible
    assert len(events.issue_464_invalid_rdate.errors) == 1
    assert events.issue_464_invalid_rdate.errors[0][0] == "RDATE"
    assert "Expected period format" in events.issue_464_invalid_rdate.errors[0][1]
    assert "199709T180000Z/PT5H30M" in events.issue_464_invalid_rdate.errors[0][1]
    assert b"RDATE:None" not in events.issue_464_invalid_rdate.to_ical()


@pytest.mark.parametrize(
    "calendar_name",
    [
        "big_bad_calendar",
        "small_bad_calendar",
        "multiple_calendar_components",
        "pr_480_summary_with_colon",
    ],
)
def test_error_message_doesnt_get_too_big(calendars, calendar_name):
    with pytest.raises(ValueError) as exception:
        calendars[calendar_name]
    # Ignore part before first : for the test.
    assert len(str(exception).split(": ", 1)[1]) <= 100



"""The RFC contains examples of conversions that we can use here."""
from icalendar import Event
from icalendar import vGeo

def test_convert_text_property_example():
    """Convert the example from Section 3.4 of :rfc:`7265`."""
    event = Event()
    event.add("SUMMARY", "Meeting with Fred")
    event.add("categories", ["Meetings", "Work"])
    jcal = event.to_jcal()
    assert jcal == [
        "vevent",
        [
            [
                "summary",
                {},
                "text",
                "Meeting with Fred",
            ],
            [
                "categories",
                {},
                "text",
                "Meetings", "Work",
            ],
        ],
        [],
    ]


def test_vGeo():
    """Check convertion of vGeo"""
    geo = vGeo((37.386013, -122.082932))
    assert geo.to_jcal("geo") == ["geo", {}, "float", [37.386013,-122.082932]]


def test_request_status(events):
    """Convert the request status example."""
    event : Event = events.rfc_7265_request_status
    jcal = event.to_jcal()
    assert jcal[1][0] == ["request-status", {}, "text", ["2.0", "Success"] ], "request-status 1"
    assert jcal[1][1] == ["request-status", {}, "text", ["3.7", "Invalid calendar user", "ATTENDEE:mailto:jsmith@example.org"]], "request-status 2"
    assert jcal == [
        "vevent",
        [   
            ["request-status", {}, "text", ["2.0", "Success"] ],
            ["request-status", {}, "text",
                [
                "3.7",
                "Invalid calendar user",
                "ATTENDEE:mailto:jsmith@example.org"
                ]
            ],
        ],
       []
    ], "whole component"



"""Tests for consolidated attribute functions in icalendar.attr module."""

from datetime import date, datetime, timedelta

import pytest

from icalendar import Event, Todo
from icalendar.attr import (
    get_duration_property,
    get_end_property,
    get_start_end_duration_with_validation,
    get_start_property,
    set_duration_with_locking,
    set_end_with_locking,
    set_start_with_locking,
)
from icalendar.error import IncompleteComponent, InvalidCalendar


class TestConsolidatedSetters:
    """Test consolidated setter functions from icalendar.attr."""

    def test_set_start_with_locking_event(self) -> None:
        """Test set_start_with_locking for Event components."""
        event = Event()
        event.add("UID", "test-event")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 14, 0))

        # Keep end locked
        set_start_with_locking(event, datetime(2026, 1, 1, 10, 0), "end", "DTEND")

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.end == datetime(2026, 1, 1, 14, 0)
        assert event.duration == timedelta(hours=4)

    def test_set_start_with_locking_todo(self) -> None:
        """Test set_start_with_locking for Todo components."""
        todo = Todo()
        todo.add("UID", "test-todo")
        todo.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        todo.add("DUE", datetime(2026, 1, 1, 14, 0))

        # Keep end locked
        set_start_with_locking(todo, datetime(2026, 1, 1, 10, 0), "end", "DUE")

        assert todo.start == datetime(2026, 1, 1, 10, 0)
        assert todo.end == datetime(2026, 1, 1, 14, 0)
        assert todo.duration == timedelta(hours=4)

    def test_set_end_with_locking_event(self) -> None:
        """Test set_end_with_locking for Event components."""
        event = Event()
        event.add("UID", "test-event")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 14, 0))

        # Keep start locked
        set_end_with_locking(event, datetime(2026, 1, 1, 16, 0), "start", "DTEND")

        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 16, 0)
        assert event.duration == timedelta(hours=4)

    def test_set_end_with_locking_todo(self) -> None:
        """Test set_end_with_locking for Todo components."""
        todo = Todo()
        todo.add("UID", "test-todo")
        todo.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        todo.add("DUE", datetime(2026, 1, 1, 14, 0))

        # Keep start locked
        set_end_with_locking(todo, datetime(2026, 1, 1, 16, 0), "start", "DUE")

        assert todo.start == datetime(2026, 1, 1, 12, 0)
        assert todo.end == datetime(2026, 1, 1, 16, 0)
        assert todo.duration == timedelta(hours=4)

    def test_set_duration_with_locking_event(self) -> None:
        """Test set_duration_with_locking for Event components."""
        event = Event()
        event.add("UID", "test-event")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 14, 0))

        # Set duration with start locked
        set_duration_with_locking(event, timedelta(hours=3), "start", "DTEND")

        assert event.start == datetime(2026, 1, 1, 12, 0)
        assert event.end == datetime(2026, 1, 1, 15, 0)
        assert event.duration == timedelta(hours=3)

    def test_set_duration_with_locking_todo(self) -> None:
        """Test set_duration_with_locking for Todo components."""
        todo = Todo()
        todo.add("UID", "test-todo")
        todo.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        todo.add("DUE", datetime(2026, 1, 1, 14, 0))

        # Set duration with start locked
        set_duration_with_locking(todo, timedelta(hours=3), "start", "DUE")

        assert todo.start == datetime(2026, 1, 1, 12, 0)
        assert todo.end == datetime(2026, 1, 1, 15, 0)
        assert todo.duration == timedelta(hours=3)

    def test_auto_detect_property_locking(self) -> None:
        """Test auto-detection of property locking."""
        event = Event()
        event.add("UID", "test-event")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DURATION", timedelta(hours=2))

        # Auto-detect should choose duration
        set_start_with_locking(
            event, datetime(2026, 1, 1, 10, 0), None, "DTEND",
        )

        assert event.start == datetime(2026, 1, 1, 10, 0)
        assert event.duration == timedelta(hours=2)
        assert event.end == datetime(2026, 1, 1, 12, 0)

    def test_error_handling_consolidated_functions(self) -> None:
        """Test error handling in consolidated functions."""
        event = Event()
        event.add("UID", "test-event")

        # Invalid locked parameter
        with pytest.raises(ValueError, match="locked must be"):
            set_start_with_locking(
                event, datetime(2026, 1, 1, 12, 0), "invalid", "DTEND",
            )

        with pytest.raises(ValueError, match="locked must be"):
            set_end_with_locking(
                event, datetime(2026, 1, 1, 12, 0), "invalid", "DTEND",
            )

        with pytest.raises(ValueError, match="locked must be"):
            set_duration_with_locking(
                event, timedelta(hours=1), "invalid", "DTEND",
            )

    def test_consolidated_functions_maintain_identical_behavior(self) -> None:
        """Test that consolidated functions maintain identical behavior."""
        # Create identical events
        event1 = Event()
        event1.add("UID", "test-event-1")
        event1.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event1.add("DTEND", datetime(2026, 1, 1, 14, 0))

        event2 = Event()
        event2.add("UID", "test-event-2")
        event2.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event2.add("DTEND", datetime(2026, 1, 1, 14, 0))

        # Use consolidated function vs direct property
        set_start_with_locking(
            event1, datetime(2026, 1, 1, 10, 0), "end", "DTEND",
        )
        event2.set_start(datetime(2026, 1, 1, 10, 0), locked="end")

        # Results should be identical
        assert event1.start == event2.start
        assert event1.end == event2.end
        assert event1.duration == event2.duration


class TestConsolidatedPropertyGetters:
    """Test consolidated property getter functions."""

    def test_get_duration_property_with_duration_set(self) -> None:
        """Test get_duration_property when DURATION is explicitly set."""
        event = Event()
        event.add("UID", "test-duration")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DURATION", timedelta(hours=1))

        duration = get_duration_property(event)
        assert duration == timedelta(hours=1)

    def test_get_duration_property_calculated(self) -> None:
        """Test get_duration_property when calculated from start/end."""
        event = Event()
        event.add("UID", "test-calculated")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 15, 0))

        duration = get_duration_property(event)
        assert duration == timedelta(hours=3)

    def test_get_start_property_success(self) -> None:
        """Test get_start_property success case."""
        event = Event()
        event.add("UID", "test-start")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))

        start = get_start_property(event)
        assert start == datetime(2026, 1, 1, 12, 0)

    def test_get_start_property_missing(self) -> None:
        """Test get_start_property when DTSTART is missing."""
        event = Event()
        event.add("UID", "test-start")

        with pytest.raises(IncompleteComponent, match="No DTSTART given"):
            get_start_property(event)

    def test_get_end_property_with_dtend(self) -> None:
        """Test get_end_property when DTEND is set."""
        event = Event()
        event.add("UID", "test-end")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 15, 0))

        end = get_end_property(event, "DTEND")
        assert end == datetime(2026, 1, 1, 15, 0)

    def test_get_end_property_with_duration(self) -> None:
        """Test get_end_property when DURATION is set."""
        event = Event()
        event.add("UID", "test-duration-end")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DURATION", timedelta(hours=3))

        end = get_end_property(event, "DTEND")
        assert end == datetime(2026, 1, 1, 15, 0)

    def test_get_end_property_default_behavior(self) -> None:
        """Test get_end_property default behavior."""
        event = Event()
        event.add("UID", "test-default")
        event.add("DTSTART", date(2026, 1, 1))

        end = get_end_property(event, "DTEND")
        assert end == date(2026, 1, 2)

        event2 = Event()
        event2.add("UID", "test-default-2")
        event2.add("DTSTART", datetime(2026, 1, 1, 12, 0))

        end2 = get_end_property(event2, "DTEND")
        assert end2 == datetime(2026, 1, 1, 12, 0)

    def test_get_start_end_duration_validation_success(self) -> None:
        """Test get_start_end_duration_with_validation success case."""
        event = Event()
        event.add("UID", "test-validation")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 15, 0))

        start, end, duration = get_start_end_duration_with_validation(
            event, "DTSTART", "DTEND", "VEVENT",
        )
        assert start == datetime(2026, 1, 1, 12, 0)
        assert end == datetime(2026, 1, 1, 15, 0)
        assert duration is None

    def test_get_start_end_duration_validation_errors(self) -> None:
        """Test get_start_end_duration_with_validation error cases."""
        # Test both DTEND and DURATION
        event = Event()
        event.add("UID", "test-both")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))
        event.add("DTEND", datetime(2026, 1, 1, 15, 0))
        event.add("DURATION", timedelta(hours=2))

        with pytest.raises(InvalidCalendar, match="Only one of DTEND and DURATION"):
            get_start_end_duration_with_validation(
                event, "DTSTART", "DTEND", "VEVENT",
            )

        # Test invalid duration for date DTSTART
        event2 = Event()
        event2.add("UID", "test-date-duration")
        event2.add("DTSTART", date(2026, 1, 1))
        event2.add("DURATION", timedelta(hours=2))

        with pytest.raises(InvalidCalendar, match="DURATION must be of days"):
            get_start_end_duration_with_validation(
                event2, "DTSTART", "DTEND", "VEVENT",
            )

        # Test type mismatch
        event3 = Event()
        event3.add("UID", "test-type-mismatch")
        event3.add("DTSTART", date(2026, 1, 1))
        event3.add("DTEND", datetime(2026, 1, 1, 15, 0))

        with pytest.raises(InvalidCalendar, match="must be of the same type"):
            get_start_end_duration_with_validation(
                event3, "DTSTART", "DTEND", "VEVENT",
            )



import unittest
from datetime import datetime

from icalendar import Calendar, cli
from icalendar.compatibility import ZoneInfo

INPUT = """
BEGIN:VCALENDAR
VERSION:2.0
CALSCALE:GREGORIAN
BEGIN:VEVENT
SUMMARY:Test Summary
ORGANIZER:organizer@test.test
ATTENDEE:attendee1@example.com
ATTENDEE:attendee2@test.test
COMMENT:Comment
DTSTART;TZID=Europe/Warsaw:20220820T103400
DTEND;TZID=Europe/Warsaw:20220820T113400
LOCATION:New Amsterdam, 1000 Sunrise Test Street
DESCRIPTION: Test Description
END:VEVENT
BEGIN:VEVENT
ORGANIZER:organizer@test.test
ATTENDEE:attendee1@example.com
SUMMARY:Test summary
DTSTART;TZID=Europe/Warsaw:20220820T200000
DTEND;TZID=Europe/Warsaw:20220820T203000
LOCATION:New Amsterdam, 1010 Test Street
DESCRIPTION:Test Description\\nThis one is multiline
END:VEVENT
BEGIN:VEVENT
UID:1
SUMMARY:TEST
DTSTART:20220511
DURATION:P5D
END:VEVENT
END:VCALENDAR
"""


def local_datetime(dt):
    return (
        datetime.strptime(dt, "%Y%m%dT%H%M%S")
        .replace(tzinfo=ZoneInfo("Europe/Warsaw"))
        .astimezone()
        .strftime("%c")
    )


# datetimes are displayed in the local timezone, so we cannot just hardcode them
firststart = local_datetime("20220820T103400")
firstend = local_datetime("20220820T113400")
secondstart = local_datetime("20220820T200000")
secondend = local_datetime("20220820T203000")

PROPER_OUTPUT = f"""    Organizer: organizer <organizer@test.test>
    Attendees:
     attendee1 <attendee1@example.com>
     attendee2 <attendee2@test.test>
    Summary    : Test Summary
    Starts     : {firststart}
    End        : {firstend}
    Duration   : 1:00:00
    Location   : New Amsterdam, 1000 Sunrise Test Street
    Comment    : Comment
    Description:
      Test Description

    Organizer: organizer <organizer@test.test>
    Attendees:
     attendee1 <attendee1@example.com>
    Summary    : Test summary
    Starts     : {secondstart}
    End        : {secondend}
    Duration   : 0:30:00
    Location   : New Amsterdam, 1010 Test Street
    Comment    : 
    Description:
     Test Description
     This one is multiline

    Organizer: 
    Attendees:

    Summary    : TEST
    Starts     : Wed May 11 00:00:00 2022
    End        : Mon May 16 00:00:00 2022
    Duration   : 5 days, 0:00:00
    Location   : 
    Comment    : 
    Description:
     

"""  # noqa: W291, W293


class CLIToolTest(unittest.TestCase):
    def test_output_is_proper(self):
        self.maxDiff = None
        calendar = Calendar.from_ical(INPUT)
        output = ""
        for event in calendar.walk("vevent"):
            output += cli.view(event) + "\n\n"
        assert output == PROPER_OUTPUT


if __name__ == "__main__":
    unittest.main()



"""Test the identity and equality between properties."""

from datetime import date, datetime, time

from icalendar import vDDDTypes
from icalendar.timezone.zoneinfo import zoneinfo

try:
    import pytz
except ImportError:
    pytz = None
from copy import deepcopy

import pytest
from dateutil import tz

vDDDTypes_list = [
    vDDDTypes(
        datetime(
            year=2022,
            month=7,
            day=22,
            hour=12,
            minute=7,
            tzinfo=zoneinfo.ZoneInfo("Europe/London"),
        )
    ),
    vDDDTypes(datetime(year=2022, month=7, day=22, hour=12, minute=7)),
    vDDDTypes(datetime(year=2022, month=7, day=22, hour=12, minute=7, tzinfo=tz.UTC)),
    vDDDTypes(date(year=2022, month=7, day=22)),
    vDDDTypes(date(year=2022, month=7, day=23)),
    vDDDTypes(time(hour=22, minute=7, second=2)),
]
if pytz:
    vDDDTypes_list.append(
        vDDDTypes(
            pytz.timezone("EST").localize(
                datetime(year=2022, month=7, day=22, hour=12, minute=7)
            )
        ),
    )


def identity(x):
    return x


@pytest.mark.parametrize(
    "map",
    [
        deepcopy,
        identity,
        hash,
    ],
)
@pytest.mark.parametrize("v_type", vDDDTypes_list)
@pytest.mark.parametrize("other", vDDDTypes_list)
def test_vDDDTypes_equivalance(map, v_type, other):
    if v_type is other:
        assert map(v_type) == map(other), f"identity implies equality: {map.__name__}()"
        assert map(v_type) == map(other), f"identity implies equality: {map.__name__}()"
    else:
        assert map(v_type) != map(other), f"expected inequality: {map.__name__}()"
        assert map(v_type) != map(other), f"expected inequality: {map.__name__}()"


@pytest.mark.parametrize("v_type", vDDDTypes_list)
def test_inequality_with_different_types(v_type):
    assert v_type != 42
    assert v_type != "test"
